"""ClickHouse data backend for Ralph."""

import json
import logging
from datetime import datetime
from io import IOBase
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
)
from uuid import UUID, uuid4

import clickhouse_connect
from clickhouse_connect.driver.client import Client
from clickhouse_connect.driver.exceptions import ClickHouseError
from pydantic import BaseModel, PositiveInt, ValidationError

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseOperationType,
    BaseQuery,
    DataBackendStatus,
    Listable,
    Writable,
)
from ralph.conf import BaseSettingsConfig, ClientOptions
from ralph.exceptions import BackendException
from ralph.utils import iter_by_batch, parse_iterable_to_dict

logger = logging.getLogger(__name__)


class ClickHouseInsert(BaseModel):
    """Model to validate required fields for ClickHouse insertion."""

    event_id: UUID
    emission_time: datetime


class ClickHouseClientOptions(ClientOptions):
    """Pydantic model for `clickhouse` client options."""

    date_time_input_format: str = "best_effort"


class InsertTuple(NamedTuple):
    """Named tuple for ClickHouse insertion."""

    event_id: UUID
    emission_time: datetime
    event: str


class ClickHouseDataBackendSettings(BaseDataBackendSettings):
    """ClickHouse data backend default configuration.

    Attributes:
        HOST (str): ClickHouse server host to connect to.
        PORT (int): ClickHouse server port to connect to.
        DATABASE (str): ClickHouse database to connect to.
        EVENT_TABLE_NAME (str): Table where events live.
        USERNAME (str): ClickHouse username to connect as (optional).
        PASSWORD (str): Password for the given ClickHouse username (optional).
        CLIENT_OPTIONS (ClickHouseClientOptions): A dictionary of valid options for the
            ClickHouse client connection.
        LOCALE_ENCODING (str): The locale encoding to use when none is provided.
        READ_CHUNK_SIZE (int): The default chunk size for reading.
        WRITE_CHUNK_SIZE (int): The default chunk size for writing.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__CLICKHOUSE__"

    HOST: str = "localhost"
    PORT: int = 8123
    DATABASE: str = "xapi"
    EVENT_TABLE_NAME: str = "xapi_events_all"
    USERNAME: Optional[str] = None
    PASSWORD: Optional[str] = None
    CLIENT_OPTIONS: ClickHouseClientOptions = ClickHouseClientOptions()


class ClickHouseQuery(BaseQuery):
    """Base ClickHouse query model.

    Attributes:
        select (str or list): Name of the table(s) to query.
        where (str or list): Where expression for filtering the data.
        parameters (dict): Dictionary of substitution values.
        limit (int): Maximum number of rows to return.
        sort (str): Order by expression determining the sorting direction.
        column_oriented (bool): Whether to return the results as a sequence of columns
            rather than a sequence of rows.
    """

    select: Union[str, List[str]] = "event"
    where: Union[str, List[str], None]
    parameters: Union[Dict, None]
    limit: Union[int, None]
    sort: Union[str, None]
    column_oriented: Union[bool, None] = False


Settings = TypeVar("Settings", bound=ClickHouseDataBackendSettings)


class ClickHouseDataBackend(
    BaseDataBackend[Settings, ClickHouseQuery],
    Writable,
    Listable,
):
    """ClickHouse database backend."""

    name = "clickhouse"
    default_operation_type = BaseOperationType.CREATE
    unsupported_operation_types = {
        BaseOperationType.APPEND,
        BaseOperationType.DELETE,
        BaseOperationType.UPDATE,
    }

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the ClickHouse configuration.

        Args:
            settings (ClickHouseDataBackendSettings or None): The ClickHouse
                data backend settings.
        """
        super().__init__(settings)
        self.database = self.settings.DATABASE
        self._client = None

    @property
    def client(self) -> Client:
        """Create a ClickHouse client if it doesn't exist.

        We do this here so that we don't interrupt initialization in the case
        where ClickHouse is not running when Ralph starts up, which will cause
        Ralph to hang. This client is HTTP, so not actually stateful. Ralph
        should be able to gracefully deal with ClickHouse outages at all other
        times.
        """
        if not self._client:
            self._client = clickhouse_connect.get_client(
                host=self.settings.HOST,
                port=self.settings.PORT,
                database=self.database,
                username=self.settings.USERNAME,
                password=self.settings.PASSWORD,
                settings=self.settings.CLIENT_OPTIONS.dict(),
            )
        return self._client

    def status(self) -> DataBackendStatus:
        """Check ClickHouse connection status.

        Return:
            DataBackendStatus: The status of the data backend.
        """
        try:
            self.client.query("SELECT 1")
        except ClickHouseError:
            return DataBackendStatus.AWAY

        return DataBackendStatus.OK

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List tables for a given database.

        Args:
            target (str): The database name to list tables from.
            details (bool): Get detailed table information instead of just table names.
            new (bool): Ignored.

        Yield:
            str: The next table name. (If `details` is False).
            dict: The next table details. (If `details` is True).

        Raise:
            BackendException: If a failure during table names retrieval occurs.
        """
        sql = f"SHOW TABLES FROM {target if target else self.database}"

        try:
            tables = self.client.query(sql).named_results()
        except (ClickHouseError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to read tables: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        if new:
            logger.warning("The `new` argument is ignored")

        for table in tables:
            if details:
                yield table
            else:
                yield str(table.get("name"))

    def read(  # noqa: PLR0913
        self,
        query: Optional[ClickHouseQuery] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read documents matching the query in the target table and yield them.

        Args:
            query (ClickHouseQuery): The query to use when fetching documents.
            target (str or None): The target table name to query.
                If target is `None`, the `EVENT_TABLE_NAME` is used instead.
            chunk_size (int or None): The chunk size when reading documents by batches.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield dictionaries or bytes.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            bytes: The next raw document if `raw_output` is True.
            dict: The next JSON parsed document if `raw_output` is False.

        Raise:
            BackendException: If a failure occurs during ClickHouse connection or
                during encoding documents and `ignore_errors` is set to `False`.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_dicts(
        self,
        query: ClickHouseQuery,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        target = target if target else self.settings.EVENT_TABLE_NAME

        if isinstance(query.select, str):
            query.select = [query.select]
        select = ",".join(query.select)
        sql = f"SELECT {select} FROM {target}"  # noqa: S608

        if query.where:
            if isinstance(query.where, str):
                query.where = [query.where]
            filter_str = "\nWHERE 1=1 AND "
            filter_str += """
            AND
            """.join(
                query.where
            )
            sql += filter_str

        if query.sort:
            sql += f"\nORDER BY {query.sort}"

        if query.limit:
            sql += f"\nLIMIT {query.limit}"

        msg = "Start reading the %s table of the %s database (chunk size: %d)"
        logger.debug(msg, target, self.database, chunk_size)
        try:
            result = self.client.query(
                sql,
                parameters=query.parameters,
                settings={"buffer_size": chunk_size},
                column_oriented=query.column_oriented,
            ).named_results()
            yield from parse_iterable_to_dict(
                result, ignore_errors, self._parse_event_json
            )
        except (ClickHouseError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to read documents: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write `data` documents to the `target` table and return their count.

        Args:
            data (Iterable or IOBase): The data containing documents to write.
            target (str or None): The target table name.
                If target is `None`, the `EVENT_TABLE_NAME` is used instead.
            chunk_size (int or None): The number of documents to write in one batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, errors during decoding, encoding and
                sending batches of documents are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.

        Return:
            int: The number of documents written.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If the `operation_type` is `APPEND`, `UPDATE`
                or `DELETE` as it is not supported.
        """
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,  # noqa: ARG002
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        count = 0
        target = target if target else self.settings.EVENT_TABLE_NAME
        msg = "Start writing to the %s table of the %s database (chunk size: %d)"
        logger.debug(msg, target, self.database, chunk_size)
        insert_tuples = self._to_insert_tuples(data, ignore_errors)
        for batch in iter_by_batch(insert_tuples, chunk_size):
            count += self._bulk_import(batch, ignore_errors, target)

        logger.info("Inserted a total of %d documents with success", count)
        return count

    def close(self) -> None:
        """Close the ClickHouse backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._client:
            logger.warning("No backend client to close.")
            return

        try:
            self.client.close()
        except ClickHouseError as error:
            msg = "Failed to close ClickHouse client: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    def _to_insert_tuples(
        self,
        data: Iterable[dict],
        ignore_errors: bool = False,
    ) -> Generator[InsertTuple, None, None]:
        """Convert `data` dictionaries to insert tuples."""
        for statement in data:
            try:
                insert = ClickHouseInsert(
                    event_id=statement.get("id", str(uuid4())),
                    emission_time=statement["timestamp"],
                )
            except (KeyError, ValidationError) as error:
                msg = "Statement %s has an invalid 'id' or 'timestamp' field"
                if ignore_errors:
                    logger.warning(msg, statement)
                    continue
                logger.error(msg, statement)
                raise BackendException(msg % statement) from error

            insert_tuple = InsertTuple(
                insert.event_id,
                insert.emission_time,
                json.dumps(statement),
            )

            yield insert_tuple

    def _bulk_import(
        self,
        batch: List[InsertTuple],
        ignore_errors: bool = False,
        event_table_name: Optional[str] = None,
    ):
        """Insert a batch of documents into the selected database table."""
        try:
            found_ids = {document.event_id for document in batch}

            if len(found_ids) != len(batch):
                raise BackendException("Duplicate IDs found in batch")

            self.client.insert(
                event_table_name,
                batch,
                column_names=["event_id", "emission_time", "event"],
                # Allow ClickHouse to buffer the insert, and wait for the
                # buffer to flush. Should be configurable, but I think these are
                # reasonable defaults.
                settings={"async_insert": 1, "wait_for_async_insert": 1},
            )
        except (ClickHouseError, BackendException) as error:
            if not ignore_errors:
                raise BackendException(*error.args) from error
            msg = "Bulk import failed for current chunk but you choose to ignore it."
            logger.warning(msg)
            # There is no current way of knowing how many rows from the batch
            # succeeded, we assume 0 here.
            return 0

        inserted_count = len(batch)
        logger.debug("Inserted %d documents chunk with success", inserted_count)

        return inserted_count

    @staticmethod
    def _parse_event_json(document: Dict[str, Any]) -> Dict[str, Any]:
        """Return the `document` with a JSON parsed `event` field."""
        if "event" in document:
            document["event"] = json.loads(document["event"])

        return document



"""Problem interaction events model definitions."""

import sys
from typing import List, Union

from pydantic import Json

from ralph.models.selector import selector

from ..browser import BaseBrowserModel
from ..server import BaseServerModel
from .fields.events import (
    EdxProblemHintDemandhintDisplayedEventField,
    EdxProblemHintFeedbackDisplayedEventField,
    ProblemCheckEventField,
    ProblemCheckFailEventField,
    ProblemRescoreEventField,
    ProblemRescoreFailEventField,
    ResetProblemEventField,
    ResetProblemFailEventField,
    SaveProblemFailEventField,
    SaveProblemSuccessEventField,
    ShowAnswerEventField,
    UIProblemResetEventField,
    UIProblemShowEventField,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class EdxProblemHintDemandhintDisplayed(BaseServerModel):
    """Pydantic model for `edx.problem.hint.demandhint_displayed` statement.

    The server emits this statement when a user requests a hint for a problem.

    Attributes:
        event (dict): See EdxProblemHintDemandhintDisplayedEventField.
        event_type (str): Consists of the value `edx.problem.hint.demandhint_displayed`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.problem.hint.demandhint_displayed"
    )

    event: EdxProblemHintDemandhintDisplayedEventField
    event_type: Literal["edx.problem.hint.demandhint_displayed"]
    page: Literal["x_module"]


class EdxProblemHintFeedbackDisplayed(BaseServerModel):
    """Pydantic model for `edx.problem.hint.feedback_displayed` statement.

    The server emits this event when a user receives a hint after answering a problem.

    Attributes:
        event (dict): See EdxProblemHintFeedbackDisplayedEventField.
        event_type (str): Consists of the value `edx.problem.hint.feedback_displayed`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.problem.hint.feedback_displayed"
    )

    event: EdxProblemHintFeedbackDisplayedEventField
    event_type: Literal["edx.problem.hint.feedback_displayed"]
    page: Literal["x_module"]


class UIProblemCheck(BaseBrowserModel):
    """Pydantic model for `problem_check` statement.

    The browser emits this event when a user checks a problem.

    Attributes:
        event (str): Consists of values of problem being checked, styled as `GET`
            parameters.
        event_type (str): Consists of the value `problem_check`.
        name (str): Consists of the value `problem_check`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_check")

    event: str
    event_type: Literal["problem_check"]
    name: Literal["problem_check"]


class ProblemCheck(BaseServerModel):
    """Pydantic model for `problem_check` statement.

    The server emits this event when a user checks a problem.

    Attributes:
        event (dict): See ProblemCheckEventField.
        event_type (str): Consists of the value `problem_check`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_check")

    event: ProblemCheckEventField
    event_type: Literal["problem_check"]
    page: Literal["x_module"]


class ProblemCheckFail(BaseServerModel):
    """Pydantic model for `problem_check_fail` statement.

    The server emits this event when a user checks a problem and a failure prevents the
    problem from being checked successfully.

    Attributes:
        event (dict): See ProblemCheckFailEventField.
        event_type (str): Consists of the value `problem_check_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_check_fail")

    event: ProblemCheckFailEventField
    event_type: Literal["problem_check_fail"]
    page: Literal["x_module"]


class UIProblemGraded(BaseBrowserModel):
    """Pydantic model for `problem_graded` statement.

    The server emits this statement each time a user clicks <kbd>Check</kbd> for a
    problem and it is graded successfully.

    Attributes:
        event (list): See ProblemGradedEventField.
        event_type (str): Consists of the value `problem_graded`.
        name (str): Consists of the value `problem_graded`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_graded")

    event: List[Union[str, Literal[None], None]]
    event_type: Literal["problem_graded"]
    name: Literal["problem_graded"]


class ProblemRescore(BaseServerModel):
    """Pydantic model for `problem_rescore` statement.

    The server emits this statement when a problem is successfully rescored.

    Attributes:
        event (dict): See ProblemRescoreEventField.
        event_type (str): Consists of the value `problem_rescore`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_rescore")

    event: ProblemRescoreEventField
    event_type: Literal["problem_rescore"]
    page: Literal["x_module"]


class ProblemRescoreFail(BaseServerModel):
    """Pydantic model for `problem_rescore_fail` statement.

    The server emits this statement when a problem cannot be successfully rescored.

    Attributes:
        event (dict): See ProblemRescoreFailEventField.
        event_type (str): Consists of the value `problem_rescore_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="problem_rescore_fail")

    event: ProblemRescoreFailEventField
    event_type: Literal["problem_rescore_fail"]
    page: Literal["x_module"]


class UIProblemReset(BaseBrowserModel):
    """Pydantic model for `problem_reset` statement.

    The browser emits problem_reset events when a user clicks <kbd>Reset</kbd> to reset
    the problem answer.

    Attributes:
        event (json): See ProblemResetEventField.
        event_type (str): Consists of the value `problem_reset`.
        name (str): Consists of the value `problem_reset`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_reset")

    event: Union[
        str,
        Json[UIProblemResetEventField],
        UIProblemResetEventField,
    ]
    event_type: Literal["problem_reset"]
    name: Literal["problem_reset"]


class UIProblemSave(BaseBrowserModel):
    """Pydantic model for `problem_save` statement.

    The browser emits this statement when a user saves a problem.

    Attributes:
        event (str): Consists of all the answers saved for the problem.
        event_type (str): Consists of the value `problem_save`.
        name (str): Consists of the value `problem_save`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_save")

    event: str
    event_type: Literal["problem_save"]
    name: Literal["problem_save"]


class UIProblemShow(BaseBrowserModel):
    """Pydantic model for `problem_show` statement.

    The browser emits this statement when the answer clicks <kbd>Show Answer</kbd> to
    show the problem answer.

    Attributes:
        event (json): See ProblemShowEventField.
        event_type (str): Consists of the value `problem_save`.
        name (str): Consists of the value `problem_save`.
    """

    __selector__ = selector(event_source="browser", event_type="problem_show")

    event: Union[
        Json[UIProblemShowEventField],
        UIProblemShowEventField,
    ]
    event_type: Literal["problem_show"]
    name: Literal["problem_show"]


class ResetProblem(BaseServerModel):
    """Pydantic model for `reset_problem` statement.

    The server emits this statement when a problem has been reset successfully.

    Attributes:
        event (dict): See ResetProblemEventField.
        event_type (str): Consists of the value `reset_problem`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="reset_problem")

    event: ResetProblemEventField
    event_type: Literal["reset_problem"]
    page: Literal["x_module"]


class ResetProblemFail(BaseServerModel):
    """Pydantic model for `reset_problem_fail` statement.

    The server emits this statement when a problem cannot be reset successfully.

    Attributes:
        event (dict): See ResetProblemFailEventField.
        event_type (str): Consists of the value `reset_problem_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="reset_problem_fail")

    event: ResetProblemFailEventField
    event_type: Literal["reset_problem_fail"]
    page: Literal["x_module"]


class SaveProblemFail(BaseServerModel):
    """Pydantic model for `save_problem_fail` statement.

    The server emits this statement when a problem cannot be saved successfully.

    Attributes:
        event (dict): See SaveProblemFailEventField.
        event_type (str): Consists of the value `save_problem_fail`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="save_problem_fail")

    event: SaveProblemFailEventField
    event_type: Literal["save_problem_fail"]
    page: Literal["x_module"]


class SaveProblemSuccess(BaseServerModel):
    """Pydantic model for `save_problem_success` statement.

    The server emits this statement when a problem is saved successfully.

    Attributes:
        event (dict): See SaveProblemSuccessEventField.
        event_type (str): Consists of the value `save_problem_success`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="save_problem_success")

    event: SaveProblemSuccessEventField
    event_type: Literal["save_problem_success"]
    page: Literal["x_module"]


class ShowAnswer(BaseServerModel):
    """Pydantic model for `showanswer` statement.

    The server emits this statement when the answer to a problem is shown.

    Attributes:
        event (dict): See ShowAnswerEventField.
        event_type (str): Consists of the value `showanswer`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(event_source="server", event_type="showanswer")

    event: ShowAnswerEventField
    event_type: Literal["showanswer"]
    page: Literal["x_module"]



"""Base xAPI `Group` definitions."""

import sys
from abc import ABC
from typing import List, Optional, Union

from pydantic import StrictStr

from ..config import BaseModelWithConfig
from .agents import BaseXapiAgent
from .ifi import (
    BaseXapiAccountIFI,
    BaseXapiMboxIFI,
    BaseXapiMboxSha1SumIFI,
    BaseXapiOpenIdIFI,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseXapiGroupCommonProperties(BaseModelWithConfig, ABC):
    """Pydantic model for core `Group` type property.

    It is defined for the Group which performed the action.

    Attributes:
        objectType (str): Consists of the value `Group`.
        name (str): Consists of the full name of the Group.
    """

    objectType: Literal["Group"]
    name: Optional[StrictStr]


class BaseXapiAnonymousGroup(BaseXapiGroupCommonProperties):
    """Pydantic model for `Group` type property.

    It is defined for Anonymous Group type.

    Attributes:
        member (list): Consist of a list of the members of this Group.
    """

    member: List[BaseXapiAgent]


class BaseXapiIdentifiedGroup(BaseXapiGroupCommonProperties):
    """Pydantic model for `Group` type property.

    It is defined for Identified Group type.

    Attributes:
        member (list): Consist of a list of the members of this Group.
    """

    member: Optional[List[BaseXapiAgent]]


class BaseXapiIdentifiedGroupWithMbox(BaseXapiIdentifiedGroup, BaseXapiMboxIFI):
    """Pydantic model for `Group` type property.

    It is defined for group type with a mailto IFI.
    """


class BaseXapiIdentifiedGroupWithMboxSha1Sum(
    BaseXapiIdentifiedGroup, BaseXapiMboxSha1SumIFI
):
    """Pydantic model for `Group` type property.

    It is defined for group type with a hash IFI.
    """


class BaseXapiIdentifiedGroupWithOpenId(BaseXapiIdentifiedGroup, BaseXapiOpenIdIFI):
    """Pydantic model for `Group` type property.

    It is defined for group type with an openID IFI.
    """


class BaseXapiIdentifiedGroupWithAccount(BaseXapiIdentifiedGroup, BaseXapiAccountIFI):
    """Pydantic model for `Group` type property.

    It is defined for group type with an account IFI.
    """


BaseXapiGroup = Union[
    BaseXapiAnonymousGroup,
    BaseXapiIdentifiedGroupWithMbox,
    BaseXapiIdentifiedGroupWithMboxSha1Sum,
    BaseXapiIdentifiedGroupWithOpenId,
    BaseXapiIdentifiedGroupWithAccount,
]



"""Server event xAPI Converter."""

from typing import Set

from ralph.models.converter import ConversionItem
from ralph.models.edx.server import Server
from ralph.models.xapi.navigation.statements import PageViewed

from .base import BaseXapiConverter


class ServerEventToPageViewed(BaseXapiConverter):
    """Convert a common edX server event to xAPI.

    Example Statement: John viewed https://www.fun-mooc.fr/ page.
    """

    __src__ = Server
    __dest__ = PageViewed

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "object__id",
                    "event_type",
                    lambda event_type: self.platform_url + event_type,
                ),
            }
        )



"""Video event fields definitions."""

import sys

from ...base import AbstractBaseEventField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class VideoBaseEventField(AbstractBaseEventField):
    """Pydantic model for video core `event` field.

    Attributes:
        code (str): Consists of the `html5` value for browser-played
            videos.
        id (str): Consists of the additional videos name if given by the
            course creators, or the system-generated hash code otherwise.
    """

    class Config:  # noqa: D106
        extra = "allow"

    code: str
    id: str


class PlayVideoEventField(VideoBaseEventField):
    """Pydantic model for `play_video`.`event` field.

    Attributes:
        currentTime (float): Consists of the time in the video at which
            the statement was emitted.
    """

    currentTime: float


class PauseVideoEventField(VideoBaseEventField):
    """Pydantic model for `pause_video`.`event`.

    Attributes:
        currentTime (float): Consists of the time in the video at which
            the statement was emitted.
    """

    currentTime: float


class SeekVideoEventField(VideoBaseEventField):
    """Pydantic model for `seek_video`.`event` field.

    Attributes:
        new_time (float): Consists of the point in time the actor changed to in a media
            object during a seek operation.
        old_time (float): Consists of the point in time the actor changed from in a
            media object during a seek operation.
        type (str): Consists of the navigational method used to change position
            within the video, either `onCaptionSeek` or `onSlideSeek` value.
    """

    new_time: float
    old_time: float
    type: str


class StopVideoEventField(VideoBaseEventField):
    """Pydantic model for `stop_video`.`event` field.

    Attributes:
        currentTime (float): Consists of the time in the video at which
            the statement was emitted.
    """

    currentTime: float


class VideoHideTranscriptEventField(VideoBaseEventField):
    """Pydantic model for `hide_transcript`.`event` field.

    Attributes:
        current_time (float): Consists of the time in the video at which
            the statement was emitted.
    """

    current_time: float


class VideoShowTranscriptEventField(VideoBaseEventField):
    """Pydantic model for `show_transcript`.`event` field.

    Attributes:
        current_time (float): Consists of the time in the video at which
            the statement was emitted.
    """

    current_time: float


class SpeedChangeVideoEventField(VideoBaseEventField):
    """Pydantic model for `speed_change_video`.`event` field.

    Attributes:
        currentTime (float): Consists of the time in the video at which
            the statement was emitted.
    """

    currentTime: float
    new_speed: Literal["0.75", "1.0", "1.25", "1.50", "2.0"]
    old_speed: Literal["0.75", "1.0", "1.25", "1.50", "2.0"]



"""xAPI statement forwarding background task."""

import logging
from functools import lru_cache
from typing import List, Literal, Union

from httpx import AsyncClient, AsyncHTTPTransport, HTTPStatusError, RequestError

from ..conf import XapiForwardingConfigurationSettings, settings

logger = logging.getLogger(__name__)


@lru_cache
def get_active_xapi_forwardings() -> List[XapiForwardingConfigurationSettings]:
    """Return a list of active xAPI forwarding configuration settings."""
    active_forwardings: List = []
    if not settings.XAPI_FORWARDINGS:
        logger.info("No xAPI forwarding configured; forwarding is disabled.")
        return active_forwardings

    for forwarding in settings.XAPI_FORWARDINGS:
        if forwarding.is_active:
            active_forwardings.append(forwarding)
            continue
        msg = (
            "The RALPH_XAPI_FORWARDINGS configuration value: '%s' is not active. "
            "It will be ignored."
        )
        logger.info(msg, forwarding)

    return active_forwardings


async def forward_xapi_statements(
    statements: Union[dict, List[dict]], method: Literal["post", "put"]
) -> None:
    """Forward xAPI statements."""
    for forwarding in get_active_xapi_forwardings():
        transport = AsyncHTTPTransport(retries=forwarding.max_retries)
        async with AsyncClient(transport=transport) as client:
            try:
                # NB: post or put
                req = await getattr(client, method)(
                    forwarding.url,
                    json=statements,
                    auth=(forwarding.basic_username, forwarding.basic_password),
                    timeout=forwarding.timeout,
                )
                req.raise_for_status()
                msg = "Forwarded %s statements to %s with success."
                if isinstance(statements, list):
                    logger.debug(msg, len(statements), forwarding.url)
                else:
                    logger.debug(msg, 1, forwarding.url)
            except (RequestError, HTTPStatusError) as error:
                logger.error("Failed to forward xAPI statements. %s", error)



"""Enrollment event models context fields definitions."""

import sys
from typing import Union

from ...base import BaseContextField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class EdxCourseEnrollmentUpgradeClickedContextField(BaseContextField):
    """Pydantic model for `edx.course.enrollment.upgrade_clicked`.`context` field.

    In addition to the common context member fields, this statement also comprises the
    `mode` context member field.

    Attributes:
        mode (str): Consists of either the `audit` or `honor` value. It identifies the
            enrollment mode when the user clicked <kbd>Challenge Yourself</kbd>.
    """

    mode: Union[Literal["audit"], Literal["honor"]]


class EdxCourseEnrollmentUpgradeSucceededContextField(BaseContextField):
    """Pydantic model for `edx.course.enrollment.upgrade.succeeded`.`context` field.

    In addition to the common context member fields, this statement also comprises the
    `mode` context member field.

    Attributes:
        mode (str): Consists of the `verified` value.
    """

    mode: Literal["verified"]



"""OVH's LDP data backend for Ralph."""

import logging
from typing import Iterator, Literal, Optional, Union

import ovh
import requests
from pydantic import PositiveInt

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    DataBackendStatus,
    Listable,
)
from ralph.backends.data.mixins import HistoryMixin
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now

logger = logging.getLogger(__name__)


class LDPDataBackendSettings(BaseDataBackendSettings):
    """OVH LDP (Log Data Platform) data backend default configuration.

    Attributes:
        APPLICATION_KEY (str): The OVH API application key (AK).
        APPLICATION_SECRET (str): The OVH API application secret (AS).
        CONSUMER_KEY (str): The OVH API consumer key (CK).
        DEFAULT_STREAM_ID (str):  The default stream identifier to query.
        ENDPOINT (str): The OVH API endpoint.
        READ_CHUNK_SIZE (str): The default chunk size for reading archives.
        REQUEST_TIMEOUT (int): HTTP request timeout in seconds.
        SERVICE_NAME (str): The default LDP account name.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__LDP__"

    APPLICATION_KEY: Optional[str] = None
    APPLICATION_SECRET: Optional[str] = None
    CONSUMER_KEY: Optional[str] = None
    DEFAULT_STREAM_ID: Optional[str] = None
    ENDPOINT: Literal[
        "ovh-eu",
        "ovh-us",
        "ovh-ca",
        "kimsufi-eu",
        "kimsufi-ca",
        "soyoustart-eu",
        "soyoustart-ca",
    ] = "ovh-eu"
    REQUEST_TIMEOUT: Optional[int] = None
    SERVICE_NAME: Optional[str] = None
    READ_CHUNK_SIZE: int = 4096


class LDPDataBackend(
    BaseDataBackend[LDPDataBackendSettings, str],
    Listable,
    HistoryMixin,
):
    """OVH LDP (Log Data Platform) data backend."""

    name = "ldp"

    def __init__(self, settings: Optional[LDPDataBackendSettings] = None):
        """Instantiate the OVH LDP client.

        Args:
            settings (LDPDataBackendSettings or None): The data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self.service_name = self.settings.SERVICE_NAME
        self.stream_id = self.settings.DEFAULT_STREAM_ID
        self.timeout = self.settings.REQUEST_TIMEOUT
        self._client = None

    @property
    def client(self) -> ovh.Client:
        """Create an ovh.Client if it doesn't exist."""
        if not self._client:
            self._client = ovh.Client(
                endpoint=self.settings.ENDPOINT,
                application_key=self.settings.APPLICATION_KEY,
                application_secret=self.settings.APPLICATION_SECRET,
                consumer_key=self.settings.CONSUMER_KEY,
            )
        return self._client

    def status(self) -> DataBackendStatus:
        """Check whether the default service_name is accessible."""
        try:
            self.client.get(self._get_archive_endpoint())
        except ovh.exceptions.APIError as error:
            logger.error("Failed to connect to the LDP: %s", error)
            return DataBackendStatus.ERROR
        except BackendParameterException:
            return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List archives for a given target stream_id.

        Args:
            target (str or None): The target stream_id where to list the archives.
                If target is `None`, the `DEFAULT_STREAM_ID` is used instead.
            details (bool): Get detailed archive information in addition to archive IDs.
            new (bool): Given the history, list only not already read archives.

        Yield:
            str: If `details` is False.
            dict: If `details` is True.

        Raise:
            BackendParameterException: If the `target` is `None` and no
                `DEFAULT_STREAM_ID` is given.
            BackendException: If a failure during retrieval of archives list occurs.
        """
        list_archives_endpoint = self._get_archive_endpoint(stream_id=target)
        logger.info("List archives endpoint: %s", list_archives_endpoint)
        logger.info("List archives details: %s", str(details))

        try:
            archives = self.client.get(list_archives_endpoint)
        except ovh.exceptions.APIError as error:
            msg = "Failed to get archives list: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        logger.info("Found %d archives", len(archives))

        if new:
            archives = set(archives) - set(self.get_command_history(self.name, "read"))
            logger.debug("New archives: %d", len(archives))

        if not details:
            for archive in archives:
                yield archive

            return

        for archive in archives:
            detail = self._details(target, archive)
            if detail:
                yield detail

    def read(  # noqa: PLR0913
        self,
        query: Optional[str] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = True,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read an archive matching the query in the target stream_id and yield it.

        Args:
            query (str or BaseQuery): The ID of the archive to read.
            target (str or None): The target stream_id containing the archives.
                If target is `None`, the `DEFAULT_STREAM_ID` is used instead.
            chunk_size (int or None): The chunk size when reading archives by batch.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Should always be set to `True`.
            ignore_errors (bool): No impact as no encoding operation is performed.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            bytes: The content of the archive matching the query.

        Raise:
            BackendException: If a failure occurs during LDP connection.
            BackendParameterException: If the `query` argument is not an archive name.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_dicts(
        self,
        query: str,  # noqa: ARG002
        target: Optional[str],  # noqa: ARG002
        chunk_size: int,  # noqa: ARG002
        ignore_errors: bool,  # noqa: ARG002
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        msg = (
            "Invalid `raw_output` value. LDP data backend doesn't support yielding "
            "dictionaries with `raw_output=False`"
        )
        logger.error(msg)
        raise BackendParameterException(msg)

    def _read_bytes(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
    ) -> Iterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        if not query:
            msg = "Invalid query. The query should be a valid archive name"
            logger.error(msg)
            raise BackendParameterException(msg)

        if not ignore_errors:
            msg = "The `ignore_errors` argument is ignored"
            logger.warning(msg)

        target = target if target else self.stream_id
        msg = "Getting archive: %s from stream: %s"
        logger.debug(msg, query, target)

        # Stream response (archive content)
        url = self._url(query)
        try:
            with requests.get(url, stream=True, timeout=self.timeout) as result:
                result.raise_for_status()
                for chunk in result.iter_content(chunk_size=chunk_size):
                    yield chunk
        except requests.exceptions.HTTPError as error:
            msg = "Failed to read archive %s: %s"
            logger.error(msg, query, error)
            raise BackendException(msg % (query, error)) from error

        # Get detailed information about the archive to fetch
        details = self._details(target, query)
        # Archive is supposed to have been fully read, add a new entry to
        # the history.
        self.append_to_history(
            {
                "backend": self.name,
                "command": "read",
                # WARNING: previously only the filename was used as the ID
                # By changing this and prepending the `target` stream_id previously
                # fetched archives will not be marked as read anymore.
                "id": f"{target}/{query}",
                "filename": details.get("filename"),
                "size": details.get("size"),
                "timestamp": now(),
            }
        )

    def close(self) -> None:
        """LDP data backend has no open connections to close. No action."""
        self._client = None
        logger.info("No open connections to close; skipping")

    def _get_archive_endpoint(self, stream_id: Optional[str] = None) -> str:
        """Return OVH's archive endpoint."""
        stream_id = stream_id if stream_id else self.stream_id
        if None in (self.service_name, stream_id):
            msg = "LDPDataBackend requires to set both service_name and stream_id"
            logger.error(msg)
            raise BackendParameterException(msg)
        return (
            f"/dbaas/logs/{self.service_name}/output/graylog/stream/{stream_id}/archive"
        )

    def _url(self, name: str) -> str:
        """Get archive absolute URL."""
        download_url_endpoint = f"{self._get_archive_endpoint()}/{name}/url"
        try:
            response = self.client.post(download_url_endpoint)
        except ovh.exceptions.APIError as error:
            msg = "Failed to get '%s' archive URL: %s"
            logger.error(msg, name, error)
            raise BackendException(msg % (name, error)) from error

        download_url = response.get("url")
        logger.debug("Temporary URL: %s", download_url)
        return download_url

    def _details(self, stream_id: str, name: str) -> Optional[dict]:
        """Return `name` archive details.

        Expected JSON response looks like:

            {
                "archiveId": "5d49d1b3-a3eb-498c-9039-6a482166f888",
                "createdAt": "2020-06-18T04:38:59.436634+02:00",
                "filename": "2020-06-16.gz",
                "md5": "01585b394be0495e38dbb60b20cb40a9",
                "retrievalDelay": 0,
                "retrievalState": "sealed",
                "sha256": "645d8e21e6fdb8aa7ffc5c[...]9ce612d06df8dcf67cb29a45ca",
                "size": 67906662,
            }
        """
        try:
            return self.client.get(f"{self._get_archive_endpoint(stream_id)}/{name}")
        except ovh.exceptions.APIError as error:
            msg = "Failed to get '%s' archive details: %s"
            logger.error(msg, name, error)
            raise BackendException(msg % (name, error)) from error



"""Websocket stream backend for Ralph."""

import logging
from typing import AsyncIterator, Optional, Union

import websockets
from pydantic import AnyUrl, PositiveInt
from websockets.http import USER_AGENT

from ralph.backends.data.base import (
    BaseAsyncDataBackend,
    BaseDataBackendSettings,
    DataBackendStatus,
)
from ralph.conf import BaseSettingsConfig, ClientOptions
from ralph.exceptions import BackendException

logger = logging.getLogger(__name__)


class WSClientOptions(ClientOptions):
    """Client options for `websockets.connection`.

    For mode details, see the
    [websockets.connection documentation](https://websockets.readthedocs.io/en/stable/reference/asyncio/client.html#websockets.client.connect)

    Attributes:
        close_timeout (float): Timeout for opening the connection in seconds.
        compression (str): Per-message compression (deflate) is activated by default.
            Setting it to `None` disables compression.
        max_size (int): Maximum size of incoming messages in bytes.
            Setting it to `None` disables the limit.
        max_queue (int): Maximum number of incoming messages in receive buffer.
            Setting it to `None` disables the limit.
        open_timeout (float): Timeout for opening the connection in seconds.
            Setting it to `None` disables the timeout.
        origin (str): Value of the `Origin` header, for servers that require it.
        ping_interval (float): Delay between keepalive pings in seconds.
            Setting it to `None` disables keepalive pings.
        ping_timeout (float): Timeout for keepalive pings in seconds.
            Setting it to `None` disables timeouts.
        read_limit (int): High-water mark of read buffer in bytes.
        user_agent_header (str): Value of  the `User-Agent` request header.
            It defaults to "Python/x.y.z websockets/X.Y".
            Setting it to `None` removes the header.
        write_limit (int): High-water mark of write buffer in bytes.
    """

    close_timeout: Optional[float] = None
    compression: Optional[str] = "deflate"
    max_size: Optional[int] = 2**20
    max_queue: Optional[int] = 2**5
    open_timeout: Optional[float] = 10
    origin: Optional[str] = None
    ping_interval: Optional[float] = 20
    ping_timeout: Optional[float] = 20
    read_limit: int = 2**16
    user_agent_header: Optional[str] = USER_AGENT
    write_limit: int = 2**16


class WSDataBackendSettings(BaseDataBackendSettings):
    """Websocket data backend default configuration.

    Attributes:
        CLIENT_OPTIONS (dict): A dictionary of valid options for the websocket client
            connection. See `WSClientOptions`.
        URI (str): The URI to connect to.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__WS__"

    CLIENT_OPTIONS: WSClientOptions = WSClientOptions()
    URI: AnyUrl


class AsyncWSDataBackend(BaseAsyncDataBackend[WSDataBackendSettings, str]):
    """Websocket stream backend."""

    name = "async_ws"

    def __init__(self, settings: Optional[WSDataBackendSettings] = None):
        """Instantiate the async websocket data backend.

        Args:
            settings (WSDataBackendSettings or None): The websocket backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self._client = None

    async def client(self) -> websockets.WebSocketClientProtocol:
        """Create a websocket client connected to `settings.URI` if it doesn't exist."""
        if not self._client:
            try:
                self._client = await websockets.connect(
                    self.settings.URI, **self.settings.CLIENT_OPTIONS.dict()
                )
            except (websockets.WebSocketException, OSError, TimeoutError) as error:
                msg = "Failed open websocket connection for %s: %s"
                logger.error(msg, self.settings.URI, error)
                raise BackendException(msg % (self.settings.URI, error)) from error
        return self._client

    async def status(self) -> DataBackendStatus:
        """Implement data backend checks (e.g. connection, cluster status).

        Return:
            DataBackendStatus: The status of the data backend.
        """
        try:
            client = await self.client()
            pong_waiter = await client.ping()
            await pong_waiter
        except BackendException:
            return DataBackendStatus.ERROR
        except (websockets.WebSocketException, RuntimeError) as error:
            logger.error("Failed to Ping %s: %s", self.settings.URI, error)
            return DataBackendStatus.AWAY
        return DataBackendStatus.OK

    async def read(  # noqa: PLR0913
        self,
        query: Optional[str] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        prefetch: Optional[PositiveInt] = None,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[AsyncIterator[bytes], AsyncIterator[dict]]:
        """Read records matching the `query` in the `target` container and yield them.

        Args:
            query (str): Ignored.
            target (str or None): Ignored.
            chunk_size (int or None): Ignored.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
                If the records are dictionaries and `raw_output` is set to `True`, they
                are encoded as JSON.
                If the records are bytes and `raw_output` is set to `False`, they are
                decoded as JSON by line.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            prefetch (int): The number of records to prefetch (queue) while yielding.
                If `prefetch` is `None` it defaults to `1` - no records are prefetched.
            max_statements (int): The maximum number of statements to yield.

        Yield:
            dict: If `raw_output` is False.
            bytes: If `raw_output` is True.

        Raise:
            BackendException: If a failure during the read operation occurs or
                during encoding records and `ignore_errors` is set to `False`.
            BackendParameterException: If a backend argument value is not valid.
        """
        statements = super().read(
            query,
            target,
            chunk_size,
            raw_output,
            ignore_errors,
            prefetch,
            max_statements,
        )
        async for statement in statements:
            yield statement

    async def _read_bytes(
        self,
        query: str,  # noqa: ARG002
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> AsyncIterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        if target or chunk_size:
            logger.warning("The `target` and `chunk_size` arguments are ignored")

        client = await self.client()
        count = 0
        try:
            while event := await client.recv():
                yield bytes(f"{event}\n", encoding=self.settings.LOCALE_ENCODING)
                count += 1
        except websockets.exceptions.ConnectionClosedOK:
            logger.info("Read %s records with success", count)
        except (websockets.WebSocketException, RuntimeError) as error:
            msg = "Failed to receive message from websocket %s: %s"
            logger.error(msg, self.settings.URI, error)
            raise BackendException(msg % (self.settings.URI, error)) from error

    async def _read_dicts(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
    ) -> AsyncIterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        statements = super()._read_dicts(query, target, chunk_size, ignore_errors)
        async for statement in statements:
            yield statement

    async def close(self) -> None:
        """Close the websocket client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._client:
            logger.warning("No backend client to close.")
            return

        client = await self.client()
        try:
            await client.close()
        except (websockets.WebSocketException, OSError, TimeoutError) as error:
            msg = "Failed to close websocket connection for %s: %s"
            logger.error(msg, self.settings.URI, error)
            raise BackendException(msg % (self.settings.URI, error)) from error



"""Async MongoDB LRS backend for Ralph."""

import logging
from typing import AsyncIterator, List, Optional

from ralph.backends.data.async_mongo import AsyncMongoDataBackend
from ralph.backends.lrs.base import (
    BaseAsyncLRSBackend,
    RalphStatementsQuery,
    StatementQueryResult,
)
from ralph.backends.lrs.mongo import MongoLRSBackend, MongoLRSBackendSettings
from ralph.exceptions import BackendException, BackendParameterException

logger = logging.getLogger(__name__)


class AsyncMongoLRSBackend(
    BaseAsyncLRSBackend[MongoLRSBackendSettings], AsyncMongoDataBackend
):
    """Async MongoDB LRS backend implementation."""

    async def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""
        query = MongoLRSBackend.get_query(params)
        try:
            mongo_response = [
                document
                async for document in self.read(
                    query=query, target=target, chunk_size=params.limit
                )
            ]
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from async MongoDB")
            raise error

        search_after = None
        if mongo_response:
            search_after = mongo_response[-1]["_id"]

        return StatementQueryResult(
            statements=[document["_source"] for document in mongo_response],
            pit_id=None,
            search_after=search_after,
        )

    async def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> AsyncIterator[dict]:
        """Yield statements with matching ids from the backend."""
        query = self.query_class(filter={"_source.id": {"$in": ids}})
        try:
            async for document in self.read(query=query, target=target):
                yield document["_source"]
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from MongoDB")
            raise error



"""Tests for Ralph cli usage strings."""

import logging

from click.testing import CliRunner

from ralph.cli import cli

test_logger = logging.getLogger("ralph")


def test_cli_auth_command_usage():
    """Test ralph auth command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["auth", "--help"])

    assert result.exit_code == 0
    assert all(
        text in result.output
        for text in [
            "Options:",
            "-u, --username TEXT",
            "-p, --password TEXT",
            "-s, --scope TEXT",
            "-t, --target TEXT",
            "-M, --agent-ifi-mbox TEXT",
            "-S, --agent-ifi-mbox-sha1sum TEXT",
            "-O, --agent-ifi-openid TEXT",
            "-A, --agent-ifi-account TEXT",
            "-N, --agent-name TEXT",
            "-w, --write-to-disk",
        ]
    )

    result = runner.invoke(cli, ["auth"])
    assert result.exit_code > 0
    assert "Error: Missing option '-u' / '--username'." in result.output


def test_cli_extract_command_usage():
    """Test ralph extract command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["extract", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -p, --parser [gelf|es]  Container format parser used to extract events\n"
        "                          [required]\n"
    ) in result.output

    result = runner.invoke(cli, ["extract"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-p' / '--parser'. Choose from:\n\tgelf,\n\tes\n"
    ) in result.output


def test_cli_validate_command_usage():
    """Test ralph validate command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["validate", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  -f, --format [edx|xapi]  Input events format to validate  [required]\n"
        "  -I, --ignore-errors      Continue validating regardless of raised errors\n"
        "  -F, --fail-on-unknown    Stop validating at first unknown event\n"
    ) in result.output

    result = runner.invoke(cli, ["validate"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-f' / '--format'. Choose from:\n\tedx,\n\txapi\n"
    ) in result.output


def test_cli_convert_command_usage():
    """Test ralph convert command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["convert", "--help"])

    assert result.exit_code == 0
    assert (
        "Options:\n"
        "  From edX to xAPI converter options: \n"
        "    -u, --uuid-namespace TEXT     The UUID namespace to use for the `ID` "
        "field\n"
        "                                  generation\n"
        "    -p, --platform-url TEXT       The `actor.account.homePage` to use in"
        " the\n"
        "                                  xAPI statements  [required]\n"
        "  -f, --from [edx]                Input events format to convert  [required]\n"
        "  -t, --to [xapi]                 Output events format  [required]\n"
        "  -I, --ignore-errors             Continue writing regardless of raised "
        "errors\n"
        "  -F, --fail-on-unknown           Stop converting at first unknown event\n"
    ) in result.output

    result = runner.invoke(cli, ["convert"])
    assert result.exit_code > 0
    assert "Error: Missing option '-p' / '--platform-url'" in result.output


def test_cli_read_command_usage():
    """Test ralph read command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["read", "--help"])

    assert result.exit_code == 0
    assert (
        "Usage: ralph read [OPTIONS] [QUERY]\n\n"
        "  Read records matching the QUERY (json or string) from a configured backend."
        "\n\n"
        "Options:\n"
        "  -b, --backend "
        "[async_es|async_lrs|async_mongo|async_ws|clickhouse|es|fs|ldp|lrs|mongo|s3|"
        "swift]\n"
        "                                  Backend  [required]\n"
        "  async_es backend: \n"
        "    --async-es-allow-yellow-status / --no-async-es-allow-yellow-status\n"
        "    --async-es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-es-default-index TEXT\n"
        "    --async-es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --async-es-locale-encoding TEXT\n"
        "    --async-es-point-in-time-keep-alive TEXT\n"
        "    --async-es-read-chunk-size INTEGER\n"
        "    --async-es-refresh-after-write TEXT\n"
        "    --async-es-write-chunk-size INTEGER\n"
        "  async_lrs backend: \n"
        "    --async-lrs-base-url TEXT\n"
        "    --async-lrs-headers KEY=VALUE,KEY=VALUE\n"
        "    --async-lrs-locale-encoding TEXT\n"
        "    --async-lrs-password TEXT\n"
        "    --async-lrs-read-chunk-size INTEGER\n"
        "    --async-lrs-statements-endpoint TEXT\n"
        "    --async-lrs-status-endpoint TEXT\n"
        "    --async-lrs-username TEXT\n"
        "    --async-lrs-write-chunk-size INTEGER\n"
        "  async_mongo backend: \n"
        "    --async-mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-mongo-connection-uri MONGODSN\n"
        "    --async-mongo-default-collection TEXT\n"
        "    --async-mongo-default-database TEXT\n"
        "    --async-mongo-locale-encoding TEXT\n"
        "    --async-mongo-read-chunk-size INTEGER\n"
        "    --async-mongo-write-chunk-size INTEGER\n"
        "  async_ws backend: \n"
        "    --async-ws-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-ws-locale-encoding TEXT\n"
        "    --async-ws-read-chunk-size INTEGER\n"
        "    --async-ws-uri TEXT\n"
        "    --async-ws-write-chunk-size INTEGER\n"
        "  clickhouse backend: \n"
        "    --clickhouse-client-options KEY=VALUE,KEY=VALUE\n"
        "    --clickhouse-database TEXT\n"
        "    --clickhouse-event-table-name TEXT\n"
        "    --clickhouse-host TEXT\n"
        "    --clickhouse-locale-encoding TEXT\n"
        "    --clickhouse-password TEXT\n"
        "    --clickhouse-port INTEGER\n"
        "    --clickhouse-read-chunk-size INTEGER\n"
        "    --clickhouse-username TEXT\n"
        "    --clickhouse-write-chunk-size INTEGER\n"
        "  es backend: \n"
        "    --es-allow-yellow-status / --no-es-allow-yellow-status\n"
        "    --es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --es-default-index TEXT\n"
        "    --es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --es-locale-encoding TEXT\n"
        "    --es-point-in-time-keep-alive TEXT\n"
        "    --es-read-chunk-size INTEGER\n"
        "    --es-refresh-after-write TEXT\n"
        "    --es-write-chunk-size INTEGER\n"
        "  fs backend: \n"
        "    --fs-default-directory-path PATH\n"
        "    --fs-default-query-string TEXT\n"
        "    --fs-locale-encoding TEXT\n"
        "    --fs-read-chunk-size INTEGER\n"
        "    --fs-write-chunk-size INTEGER\n"
        "  ldp backend: \n"
        "    --ldp-application-key TEXT\n"
        "    --ldp-application-secret TEXT\n"
        "    --ldp-consumer-key TEXT\n"
        "    --ldp-default-stream-id TEXT\n"
        "    --ldp-endpoint TEXT\n"
        "    --ldp-locale-encoding TEXT\n"
        "    --ldp-read-chunk-size INTEGER\n"
        "    --ldp-request-timeout TEXT\n"
        "    --ldp-service-name TEXT\n"
        "    --ldp-write-chunk-size INTEGER\n"
        "  lrs backend: \n"
        "    --lrs-base-url TEXT\n"
        "    --lrs-headers KEY=VALUE,KEY=VALUE\n"
        "    --lrs-locale-encoding TEXT\n"
        "    --lrs-password TEXT\n"
        "    --lrs-read-chunk-size INTEGER\n"
        "    --lrs-statements-endpoint TEXT\n"
        "    --lrs-status-endpoint TEXT\n"
        "    --lrs-username TEXT\n"
        "    --lrs-write-chunk-size INTEGER\n"
        "  mongo backend: \n"
        "    --mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --mongo-connection-uri MONGODSN\n"
        "    --mongo-default-collection TEXT\n"
        "    --mongo-default-database TEXT\n"
        "    --mongo-locale-encoding TEXT\n"
        "    --mongo-read-chunk-size INTEGER\n"
        "    --mongo-write-chunk-size INTEGER\n"
        "  s3 backend: \n"
        "    --s3-access-key-id TEXT\n"
        "    --s3-default-bucket-name TEXT\n"
        "    --s3-default-region TEXT\n"
        "    --s3-endpoint-url TEXT\n"
        "    --s3-locale-encoding TEXT\n"
        "    --s3-read-chunk-size INTEGER\n"
        "    --s3-secret-access-key TEXT\n"
        "    --s3-session-token TEXT\n"
        "    --s3-write-chunk-size INTEGER\n"
        "  swift backend: \n"
        "    --swift-auth-url TEXT\n"
        "    --swift-default-container TEXT\n"
        "    --swift-identity-api-version TEXT\n"
        "    --swift-locale-encoding TEXT\n"
        "    --swift-object-storage-url TEXT\n"
        "    --swift-password TEXT\n"
        "    --swift-project-domain-name TEXT\n"
        "    --swift-read-chunk-size INTEGER\n"
        "    --swift-region-name TEXT\n"
        "    --swift-tenant-id TEXT\n"
        "    --swift-tenant-name TEXT\n"
        "    --swift-username TEXT\n"
        "    --swift-user-domain-name TEXT\n"
        "    --swift-write-chunk-size INTEGER\n"
        "  -s, --chunk-size INTEGER        Get events by chunks of size #\n"
        "  -t, --target TEXT               Endpoint from which to read events (e.g.\n"
        "                                  `/statements`)\n"
        "  -i, --ignore_errors BOOLEAN     Ignore errors during the encoding operation."
        "\n"
        "                                  [default: False]\n"
        "  --help                          Show this message and exit.\n"
    ) == result.output
    logging.warning(result.output)
    result = runner.invoke(cli, ["read"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-b' / '--backend'. "
        "Choose from:\n\tasync_es,\n\tasync_lrs,\n\tasync_mongo,\n\tasync_ws,"
        "\n\tclickhouse,\n\tes,\n\tfs,\n\tldp,\n\tlrs,\n\tmongo,\n\ts3,\n\tswift\n"
    ) in result.output


def test_cli_list_command_usage():
    """Test ralph list command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["list", "--help"])

    assert result.exit_code == 0
    assert (
        "Usage: ralph list [OPTIONS]\n\n"
        "  List available documents from a configured data backend.\n\n"
        "Options:\n"
        "  -b, --backend [async_es|async_mongo|clickhouse|es|fs|ldp|mongo|s3|swift]\n"
        "                                  Backend  [required]\n"
        "  async_es backend: \n"
        "    --async-es-allow-yellow-status / --no-async-es-allow-yellow-status\n"
        "    --async-es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-es-default-index TEXT\n"
        "    --async-es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --async-es-locale-encoding TEXT\n"
        "    --async-es-point-in-time-keep-alive TEXT\n"
        "    --async-es-read-chunk-size INTEGER\n"
        "    --async-es-refresh-after-write TEXT\n"
        "    --async-es-write-chunk-size INTEGER\n"
        "  async_mongo backend: \n"
        "    --async-mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-mongo-connection-uri MONGODSN\n"
        "    --async-mongo-default-collection TEXT\n"
        "    --async-mongo-default-database TEXT\n"
        "    --async-mongo-locale-encoding TEXT\n"
        "    --async-mongo-read-chunk-size INTEGER\n"
        "    --async-mongo-write-chunk-size INTEGER\n"
        "  clickhouse backend: \n"
        "    --clickhouse-client-options KEY=VALUE,KEY=VALUE\n"
        "    --clickhouse-database TEXT\n"
        "    --clickhouse-event-table-name TEXT\n"
        "    --clickhouse-host TEXT\n"
        "    --clickhouse-locale-encoding TEXT\n"
        "    --clickhouse-password TEXT\n"
        "    --clickhouse-port INTEGER\n"
        "    --clickhouse-read-chunk-size INTEGER\n"
        "    --clickhouse-username TEXT\n"
        "    --clickhouse-write-chunk-size INTEGER\n"
        "  es backend: \n"
        "    --es-allow-yellow-status / --no-es-allow-yellow-status\n"
        "    --es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --es-default-index TEXT\n"
        "    --es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --es-locale-encoding TEXT\n"
        "    --es-point-in-time-keep-alive TEXT\n"
        "    --es-read-chunk-size INTEGER\n"
        "    --es-refresh-after-write TEXT\n"
        "    --es-write-chunk-size INTEGER\n"
        "  fs backend: \n"
        "    --fs-default-directory-path PATH\n"
        "    --fs-default-query-string TEXT\n"
        "    --fs-locale-encoding TEXT\n"
        "    --fs-read-chunk-size INTEGER\n"
        "    --fs-write-chunk-size INTEGER\n"
        "  ldp backend: \n"
        "    --ldp-application-key TEXT\n"
        "    --ldp-application-secret TEXT\n"
        "    --ldp-consumer-key TEXT\n"
        "    --ldp-default-stream-id TEXT\n"
        "    --ldp-endpoint TEXT\n"
        "    --ldp-locale-encoding TEXT\n"
        "    --ldp-read-chunk-size INTEGER\n"
        "    --ldp-request-timeout TEXT\n"
        "    --ldp-service-name TEXT\n"
        "    --ldp-write-chunk-size INTEGER\n"
        "  mongo backend: \n"
        "    --mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --mongo-connection-uri MONGODSN\n"
        "    --mongo-default-collection TEXT\n"
        "    --mongo-default-database TEXT\n"
        "    --mongo-locale-encoding TEXT\n"
        "    --mongo-read-chunk-size INTEGER\n"
        "    --mongo-write-chunk-size INTEGER\n"
        "  s3 backend: \n"
        "    --s3-access-key-id TEXT\n"
        "    --s3-default-bucket-name TEXT\n"
        "    --s3-default-region TEXT\n"
        "    --s3-endpoint-url TEXT\n"
        "    --s3-locale-encoding TEXT\n"
        "    --s3-read-chunk-size INTEGER\n"
        "    --s3-secret-access-key TEXT\n"
        "    --s3-session-token TEXT\n"
        "    --s3-write-chunk-size INTEGER\n"
        "  swift backend: \n"
        "    --swift-auth-url TEXT\n"
        "    --swift-default-container TEXT\n"
        "    --swift-identity-api-version TEXT\n"
        "    --swift-locale-encoding TEXT\n"
        "    --swift-object-storage-url TEXT\n"
        "    --swift-password TEXT\n"
        "    --swift-project-domain-name TEXT\n"
        "    --swift-read-chunk-size INTEGER\n"
        "    --swift-region-name TEXT\n"
        "    --swift-tenant-id TEXT\n"
        "    --swift-tenant-name TEXT\n"
        "    --swift-username TEXT\n"
        "    --swift-user-domain-name TEXT\n"
        "    --swift-write-chunk-size INTEGER\n"
        "  -t, --target TEXT               Container to list events from\n"
        "  -n, --new / -a, --all           List not fetched (or all) documents\n"
        "  -D, --details / -I, --ids       Get documents detailed output (JSON)\n"
        "  --help                          Show this message and exit.\n"
    ) == result.output

    result = runner.invoke(cli, ["list"])
    assert result.exit_code > 0
    assert (
        "Error: Missing option '-b' / '--backend'. Choose from:\n\tasync_es,\n\t"
        "async_mongo,\n\tclickhouse,\n\tes,\n\tfs,\n\tldp,\n\tmongo,\n\ts3,"
        "\n\tswift\n"
    ) in result.output


def test_cli_write_command_usage():
    """Test ralph write command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["write", "--help"])

    assert result.exit_code == 0

    expected_output = (
        "Usage: ralph write [OPTIONS]\n\n"
        "  Write an archive to a configured backend.\n\n"
        "Options:\n"
        "  -b, --backend "
        "[async_es|async_lrs|async_mongo|clickhouse|es|fs|lrs|mongo|s3|swift]"
        "\n"
        "                                  Backend  [required]\n"
        "  async_es backend: \n"
        "    --async-es-allow-yellow-status / --no-async-es-allow-yellow-status\n"
        "    --async-es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-es-default-index TEXT\n"
        "    --async-es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --async-es-locale-encoding TEXT\n"
        "    --async-es-point-in-time-keep-alive TEXT\n"
        "    --async-es-read-chunk-size INTEGER\n"
        "    --async-es-refresh-after-write TEXT\n"
        "    --async-es-write-chunk-size INTEGER\n"
        "  async_lrs backend: \n"
        "    --async-lrs-base-url TEXT\n"
        "    --async-lrs-headers KEY=VALUE,KEY=VALUE\n"
        "    --async-lrs-locale-encoding TEXT\n"
        "    --async-lrs-password TEXT\n"
        "    --async-lrs-read-chunk-size INTEGER\n"
        "    --async-lrs-statements-endpoint TEXT\n"
        "    --async-lrs-status-endpoint TEXT\n"
        "    --async-lrs-username TEXT\n"
        "    --async-lrs-write-chunk-size INTEGER\n"
        "  async_mongo backend: \n"
        "    --async-mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-mongo-connection-uri MONGODSN\n"
        "    --async-mongo-default-collection TEXT\n"
        "    --async-mongo-default-database TEXT\n"
        "    --async-mongo-locale-encoding TEXT\n"
        "    --async-mongo-read-chunk-size INTEGER\n"
        "    --async-mongo-write-chunk-size INTEGER\n"
        "  clickhouse backend: \n"
        "    --clickhouse-client-options KEY=VALUE,KEY=VALUE\n"
        "    --clickhouse-database TEXT\n"
        "    --clickhouse-event-table-name TEXT\n"
        "    --clickhouse-host TEXT\n"
        "    --clickhouse-locale-encoding TEXT\n"
        "    --clickhouse-password TEXT\n"
        "    --clickhouse-port INTEGER\n"
        "    --clickhouse-read-chunk-size INTEGER\n"
        "    --clickhouse-username TEXT\n"
        "    --clickhouse-write-chunk-size INTEGER\n"
        "  es backend: \n"
        "    --es-allow-yellow-status / --no-es-allow-yellow-status\n"
        "    --es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --es-default-index TEXT\n"
        "    --es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --es-locale-encoding TEXT\n"
        "    --es-point-in-time-keep-alive TEXT\n"
        "    --es-read-chunk-size INTEGER\n"
        "    --es-refresh-after-write TEXT\n"
        "    --es-write-chunk-size INTEGER\n"
        "  fs backend: \n"
        "    --fs-default-directory-path PATH\n"
        "    --fs-default-query-string TEXT\n"
        "    --fs-locale-encoding TEXT\n"
        "    --fs-read-chunk-size INTEGER\n"
        "    --fs-write-chunk-size INTEGER\n"
        "  lrs backend: \n"
        "    --lrs-base-url TEXT\n"
        "    --lrs-headers KEY=VALUE,KEY=VALUE\n"
        "    --lrs-locale-encoding TEXT\n"
        "    --lrs-password TEXT\n"
        "    --lrs-read-chunk-size INTEGER\n"
        "    --lrs-statements-endpoint TEXT\n"
        "    --lrs-status-endpoint TEXT\n"
        "    --lrs-username TEXT\n"
        "    --lrs-write-chunk-size INTEGER\n"
        "  mongo backend: \n"
        "    --mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --mongo-connection-uri MONGODSN\n"
        "    --mongo-default-collection TEXT\n"
        "    --mongo-default-database TEXT\n"
        "    --mongo-locale-encoding TEXT\n"
        "    --mongo-read-chunk-size INTEGER\n"
        "    --mongo-write-chunk-size INTEGER\n"
        "  s3 backend: \n"
        "    --s3-access-key-id TEXT\n"
        "    --s3-default-bucket-name TEXT\n"
        "    --s3-default-region TEXT\n"
        "    --s3-endpoint-url TEXT\n"
        "    --s3-locale-encoding TEXT\n"
        "    --s3-read-chunk-size INTEGER\n"
        "    --s3-secret-access-key TEXT\n"
        "    --s3-session-token TEXT\n"
        "    --s3-write-chunk-size INTEGER\n"
        "  swift backend: \n"
        "    --swift-auth-url TEXT\n"
        "    --swift-default-container TEXT\n"
        "    --swift-identity-api-version TEXT\n"
        "    --swift-locale-encoding TEXT\n"
        "    --swift-object-storage-url TEXT\n"
        "    --swift-password TEXT\n"
        "    --swift-project-domain-name TEXT\n"
        "    --swift-read-chunk-size INTEGER\n"
        "    --swift-region-name TEXT\n"
        "    --swift-tenant-id TEXT\n"
        "    --swift-tenant-name TEXT\n"
        "    --swift-username TEXT\n"
        "    --swift-user-domain-name TEXT\n"
        "    --swift-write-chunk-size INTEGER\n"
        "  -t, --target TEXT               The target container to write into\n"
        "  -s, --chunk-size INTEGER        Get events by chunks of size #\n"
        "  -I, --ignore-errors             Continue writing regardless of raised errors"
        "\n"
        "  -o, --operation-type OP_TYPE    Either index, create, delete, update or "
        "append\n"
        "  -c, --concurrency INTEGER       Number of chunks to write concurrently. ("
        "async\n"
        "                                  backends only)\n"
        "  --help                          Show this message and exit.\n"
    )
    assert expected_output == result.output

    result = runner.invoke(cli, ["write"])
    assert result.exit_code > 0
    assert (
        "Missing option '-b' / '--backend'. Choose from:\n\tasync_es,\n\tasync_lrs,\n\t"
        "async_mongo,\n\tclickhouse,\n\tes,\n\tfs,\n\tlrs,\n\tmongo,\n\ts3,\n\tswift\n"
    ) in result.output


def test_cli_runserver_command_usage():
    """Test ralph runserver command usage."""
    runner = CliRunner()
    result = runner.invoke(cli, ["runserver", "--help"])

    expected_output = (
        "Usage: ralph runserver [OPTIONS]\n\n"
        "  Run the API server for the development environment.\n\n"
        "  Starts uvicorn programmatically for convenience and documentation.\n\n"
        "Options:\n"
        "  -b, --backend [async_es|async_mongo|clickhouse|es|fs|mongo]\n"
        "                                  Backend  [required]\n"
        "  async_es backend: \n"
        "    --async-es-allow-yellow-status / --no-async-es-allow-yellow-status\n"
        "    --async-es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-es-default-index TEXT\n"
        "    --async-es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --async-es-locale-encoding TEXT\n"
        "    --async-es-point-in-time-keep-alive TEXT\n"
        "    --async-es-read-chunk-size INTEGER\n"
        "    --async-es-refresh-after-write TEXT\n"
        "    --async-es-write-chunk-size INTEGER\n"
        "  async_mongo backend: \n"
        "    --async-mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --async-mongo-connection-uri MONGODSN\n"
        "    --async-mongo-default-collection TEXT\n"
        "    --async-mongo-default-database TEXT\n"
        "    --async-mongo-locale-encoding TEXT\n"
        "    --async-mongo-read-chunk-size INTEGER\n"
        "    --async-mongo-write-chunk-size INTEGER\n"
        "  clickhouse backend: \n"
        "    --clickhouse-client-options KEY=VALUE,KEY=VALUE\n"
        "    --clickhouse-database TEXT\n"
        "    --clickhouse-event-table-name TEXT\n"
        "    --clickhouse-host TEXT\n"
        "    --clickhouse-ids-chunk-size INTEGER\n"
        "    --clickhouse-locale-encoding TEXT\n"
        "    --clickhouse-password TEXT\n"
        "    --clickhouse-port INTEGER\n"
        "    --clickhouse-read-chunk-size INTEGER\n"
        "    --clickhouse-username TEXT\n"
        "    --clickhouse-write-chunk-size INTEGER\n"
        "  es backend: \n"
        "    --es-allow-yellow-status / --no-es-allow-yellow-status\n"
        "    --es-client-options KEY=VALUE,KEY=VALUE\n"
        "    --es-default-index TEXT\n"
        "    --es-hosts VALUE1,VALUE2,VALUE3\n"
        "    --es-locale-encoding TEXT\n"
        "    --es-point-in-time-keep-alive TEXT\n"
        "    --es-read-chunk-size INTEGER\n"
        "    --es-refresh-after-write TEXT\n"
        "    --es-write-chunk-size INTEGER\n"
        "  fs backend: \n"
        "    --fs-default-directory-path PATH\n"
        "    --fs-default-lrs-file TEXT\n"
        "    --fs-default-query-string TEXT\n"
        "    --fs-locale-encoding TEXT\n"
        "    --fs-read-chunk-size INTEGER\n"
        "    --fs-write-chunk-size INTEGER\n"
        "  mongo backend: \n"
        "    --mongo-client-options KEY=VALUE,KEY=VALUE\n"
        "    --mongo-connection-uri MONGODSN\n"
        "    --mongo-default-collection TEXT\n"
        "    --mongo-default-database TEXT\n"
        "    --mongo-locale-encoding TEXT\n"
        "    --mongo-read-chunk-size INTEGER\n"
        "    --mongo-write-chunk-size INTEGER\n"
        "  -h, --host TEXT                 LRS server host name\n"
        "  -p, --port INTEGER              LRS server port\n"
        "  --help                          Show this message and exit.\n"
    )
    assert result.exit_code == 0
    assert expected_output in result.output

    result = runner.invoke(cli, ["runserver"])
    assert result.exit_code > 0
    assert (
        "Missing option '-b' / '--backend'. Choose from:\n\tasync_es,\n\tasync_mongo,\n"
        "\tclickhouse,\n\tes,\n\tfs,\n\tmongo\n"
    ) in result.output



"""Base event model definitions."""

import sys
from datetime import datetime
from ipaddress import IPv4Address
from pathlib import Path
from typing import Dict, Optional, Union

from pydantic import AnyHttpUrl, BaseModel, constr

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseModelWithConfig(BaseModel):
    """Pydantic model for base configuration shared among all models."""

    class Config:  # noqa: D106
        extra = "forbid"


class ContextModuleField(BaseModelWithConfig):
    """Pydantic model for `context`.`module` field.

    Attributes:
        usage_key (str): Consists of a block ID of the current component.
        display_name (str): Consists of a short description or title of the component.
    """

    usage_key: constr(regex=r"^block-v1:.+\+.+\+.+type@.+@[a-f0-9]{32}$")
    display_name: str
    original_usage_key: Optional[
        constr(regex=r"^block-v1:.+\+.+\+.+type@problem\+block@[a-f0-9]{32}$")
    ]
    original_usage_version: Optional[str]


class BaseContextField(BaseModelWithConfig):
    """Pydantic model for core `context` field.

    Attributes:
        course_user_tags (Dict of str): Content from `user_api_usercoursetag` table.
            Retrieved with::
                `dict(
                    UserCourseTag.objects.filter(
                        user=request.user.pk, course_id=course_key
                    ).values_list('key', 'value')
                )`
            Note::
                Is only present when a course page is requested.
                Is an empty dictionary when the user is not logged in or not found in
                the `user_api_usercoursetag` table.
        user_id (int or str or None): Consists of the ID of the authenticated user.
            Retrieved with::
                `request.user.pk` querying the `auth_user` table.
            Note::
                Is an integer when the user is found in the `auth_user` table.
                Is an empty string when an exception is raised while retrieving the ID.
                Is `None` when the user is not logged in.
        org_id (str): Consists of the organization name that lists the course.
            Retrieved with::
                `course_id.org` where `course_id` is an
                `opaque_keys.edx.locator.CourseLocator` which is created using the URL
                of the requested page.
            Note::
                Is an empty string when the requested page is not a course page.
        course_id (str): Consists of the unique identifier for the visited course page.
            Retrieved with::
                `course_id.to_deprecated_string()` where `course_id` is an
                `opaque_keys.edx.locator.CourseLocator` which is created using the URL
                of the requested page.
            Note::
                Is an empty string when the requested page is not a course page.
        path (Path): Consist of the relative URL (without the hostname) of the
            requested page.
            Retrieved with::
                `request.META['PATH_INFO']`
    """

    course_id: constr(regex=r"^$|^course-v1:.+\+.+\+.+$")
    course_user_tags: Optional[Dict[str, str]]
    module: Optional[ContextModuleField]
    org_id: str
    path: Path
    user_id: Union[int, Literal[""], None]


class AbstractBaseEventField(BaseModelWithConfig):
    """Pydantic model for core `event` fields.

    The base model does not have any attributes as event field does not have common
    sub-fields.
    """


class BaseEdxModel(BaseModelWithConfig):
    """Pydantic model for core statement.

    WARNING: it does not define the `event`, `event_type` and `event_source` fields.

    Attributes:
        username (str): Consists of the unique username identifying the logged-in user.
            Retrieved with::
                `request.user.username` querying the `auth_user` table.
            Note::
                Is an empty string when the user is not logged in.
                If an exception is raised when retrieving the username from the table
                then the value is `anonymous`.
                Usernames are made of 2-30 ASCII letters / numbers / underscores (_) /
                hyphens (-)
        ip (IPv4Address or str): Consists of the public IPv4 address of the user.
            Retrieved with::
                `get_ip(request)` cf. https://github.com/un33k/django-ipware/tree/1.1.0
            Note::
                Can be an empty string if the IP address is not found.
        agent (str): Consists of the `User-Agent` HTTP request header.
            Retrieved with::
                `request.META[HTTP_USER_AGENT]`
            Note::
                Can be an empty string if the header is not present in the request.
                Contains information about:
                    Browser name and version
                    Operating system name and version
                    Default language
        host (str): Consists of the hostname of the server.
            Retrieved with::
                `request.META[SERVER_NAME]`
        referer (Path): Consists of the `Referer` HTTP request header.
            Retrieved with::
                `request.META[HTTP_REFERER]`
            Note::
                Can be an empty string if the header is not present in the request.
                Contains the referring URL (previous URL visited by the user).
        accept_language (str): Consists of the `Accept-Language` HTTP request header.
            Retrieved with::
                `request.META[HTTP_ACCEPT_LANGUAGE]`
            Note::
                Can be an empty string if the header is not present in the request.
                Contains the default language settings of the user.
        context (BaseContextField): see BaseContextField.
        time (datetime): Consists of the UTC time in ISO format at which the event was
            emitted.
            Retrieved with::
                `datetime.datetime.utcnow()`
        page (None): Consists of the value `None`
            Note::
                In JSON the value is `null` instead of `None`.
    """

    username: Union[constr(min_length=2, max_length=30), Literal[""]]
    ip: Union[IPv4Address, Literal[""]]
    agent: str
    host: str
    referer: Union[AnyHttpUrl, Literal[""]]
    accept_language: str
    context: BaseContextField
    time: datetime
    page: None



"""Base LRS backend for Ralph."""

from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Any, AsyncIterator, Iterator, List, Literal, Optional, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, Field, NonNegativeInt

from ralph.backends.data.base import (
    BaseAsyncDataBackend,
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseQuery,
)
from ralph.conf import BaseSettingsConfig
from ralph.models.xapi.base.agents import BaseXapiAgent
from ralph.models.xapi.base.common import IRI
from ralph.models.xapi.base.groups import BaseXapiGroup


class BaseLRSBackendSettings(BaseDataBackendSettings):
    """LRS backend default configuration."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__LRS__"


@dataclass
class StatementQueryResult:
    """Result of an LRS statements query."""

    statements: List[dict]
    pit_id: Optional[str]
    search_after: Optional[str]


class IsoDatetimeStr(str):
    """ISO 8601 date time string field type."""

    @classmethod
    def __get_validators__(cls):
        """Return expected validators for the custom field."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Value is expected to be an ISO 8601 date time string.

        Note that we also accept datetime python instance that will be converted
        to an ISO 8601 date time string.
        """
        if not isinstance(v, (str, datetime)):
            raise TypeError("a string or datetime is required")

        if isinstance(v, datetime):
            return cls(v.isoformat())

        # Validate iso-string
        try:
            datetime.fromisoformat(v)
        except ValueError as err:
            raise ValueError("invalid ISO 8601 date time string") from err
        return cls(v)


class LRSStatementsQuery(BaseQuery):
    """Pydantic model for LRS query on Statements resource query parameters.

    LRS Specification:
    https://github.com/adlnet/xAPI-Spec/blob/1.0.3/xAPI-Communication.md#213-get-statements
    """

    statement_id: Optional[str] = Field(None, alias="statementId")
    voided_statement_id: Optional[str] = Field(None, alias="voidedStatementId")
    agent: Optional[Union[BaseXapiAgent, BaseXapiGroup]]
    verb: Optional[IRI]
    activity: Optional[IRI]
    registration: Optional[UUID]
    related_activities: Optional[bool] = False
    related_agents: Optional[bool] = False
    since: Optional[IsoDatetimeStr]
    until: Optional[IsoDatetimeStr]
    limit: Optional[NonNegativeInt] = 0
    format: Optional[Literal["ids", "exact", "canonical"]] = "exact"
    attachments: Optional[bool] = False
    ascending: Optional[bool] = False


class AgentParameters(BaseModel):
    """LRS query parameters for query on type Agent.

    NB: Agent refers to the data structure, NOT to the LRS query parameter.
    """

    mbox: Optional[str]
    mbox_sha1sum: Optional[str]
    openid: Optional[str]
    account__name: Optional[str]
    account__home_page: Optional[str]


class RalphStatementsQuery(LRSStatementsQuery):
    """Represents a dictionary of possible LRS query parameters."""

    agent: Optional[AgentParameters] = AgentParameters.construct()
    search_after: Optional[str]
    pit_id: Optional[str]
    authority: Optional[AgentParameters] = AgentParameters.construct()
    ignore_order: Optional[bool]


Settings = TypeVar("Settings", bound=BaseLRSBackendSettings)


class BaseLRSBackend(BaseDataBackend[Settings, Any]):
    """Base LRS backend interface."""

    @abstractmethod
    def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""

    @abstractmethod
    def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> Iterator[dict]:
        """Yield statements with matching ids from the backend."""


class BaseAsyncLRSBackend(BaseAsyncDataBackend[Settings, Any]):
    """Base async LRS backend interface."""

    @abstractmethod
    async def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""

    @abstractmethod
    async def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> AsyncIterator[dict]:
        """Return the list of matching statement IDs from the database."""



"""Constants for `Video` xAPI profile."""

# Context extensions
CONTEXT_EXTENSION_CC_SUBTITLE_LANG = (
    "https://w3id.org/xapi/video/extensions/cc-subtitle-lang"
)
CONTEXT_EXTENSION_CC_ENABLED = "https://w3id.org/xapi/video/extensions/cc-enabled"
CONTEXT_EXTENSION_COMPLETION_THRESHOLD = (
    "https://w3id.org/xapi/video/extensions/completion-threshold"
)
CONTEXT_EXTENSION_FRAME_RATE = "https://w3id.org/xapi/video/extensions/frame-rate"
CONTEXT_EXTENSION_FULL_SCREEN = "https://w3id.org/xapi/video/extensions/full-screen"
CONTEXT_EXTENSION_LENGTH = "https://w3id.org/xapi/video/extensions/length"
CONTEXT_EXTENSION_PLAYED_SEGMENTS = (
    "https://w3id.org/xapi/video/extensions/played-segments"
)
CONTEXT_EXTENSION_QUALITY = "https://w3id.org/xapi/video/extensions/quality"
CONTEXT_EXTENSION_SCREEN_SIZE = "https://w3id.org/xapi/video/extensions/screen-size"
CONTEXT_EXTENSION_SESSION_ID = "https://w3id.org/xapi/video/extensions/session-id"
CONTEXT_EXTENSION_SPEED = "https://w3id.org/xapi/video/extensions/speed"

CONTEXT_EXTENSION_TRACK = "https://w3id.org/xapi/video/extensions/track"
CONTEXT_EXTENSION_USER_AGENT = "https://w3id.org/xapi/video/extensions/user-agent"
CONTEXT_EXTENSION_VIDEO_PLAYBACK_SIZE = (
    "https://w3id.org/xapi/video/extensions/video-playback-size"
)
CONTEXT_EXTENSION_VOLUME = "https://w3id.org/xapi/video/extensions/volume"

# Result extensions
RESULT_EXTENSION_TIME = "https://w3id.org/xapi/video/extensions/time"
RESULT_EXTENSION_TIME_FROM = "https://w3id.org/xapi/video/extensions/time-from"
RESULT_EXTENSION_TIME_TO = "https://w3id.org/xapi/video/extensions/time-to"
RESULT_EXTENSION_PROGRESS = "https://w3id.org/xapi/video/extensions/progress"



"""Base xAPI `Statement` definitions."""

from datetime import datetime
from typing import Any, List, Optional, Union
from uuid import UUID

from pydantic import constr, root_validator

from ..config import BaseModelWithConfig
from .agents import BaseXapiAgent
from .attachments import BaseXapiAttachment
from .contexts import BaseXapiContext
from .groups import BaseXapiGroup
from .objects import BaseXapiObject
from .results import BaseXapiResult
from .verbs import BaseXapiVerb


class BaseXapiStatement(BaseModelWithConfig):
    """Pydantic model for base xAPI statements.

    Attributes:
        id (UUID): Consists of a generated UUID string from the source event string.
        actor (dict): Consists of a definition of who performed the action.
        verb (dict): Consists of the action between an Actor and an Activity.
        object (dict): Consists of a definition of the thing that was acted on.
        result (dict): Consists of the outcome related to the Statement.
        context (dict): Consists of contextual information for the Statement.
        timestamp (datetime): Consists of the timestamp of when the event occurred.
        stored (datetime): Consists of the timestamp of when the event was recorded.
        authority (dict): Consists of the Actor asserting this Statement is true.
        version (str): Consists of the associated xAPI version of the Statement.
        attachments (list): Consists of a list of attachments.
    """

    id: Optional[UUID]
    actor: Union[BaseXapiAgent, BaseXapiGroup]
    verb: BaseXapiVerb
    object: BaseXapiObject
    result: Optional[BaseXapiResult]
    context: Optional[BaseXapiContext]
    timestamp: Optional[datetime]
    stored: Optional[datetime]
    authority: Optional[Union[BaseXapiAgent, BaseXapiGroup]]
    version: constr(regex=r"^1\.0\.[0-9]+$") = "1.0.0"
    attachments: Optional[List[BaseXapiAttachment]]

    @root_validator(pre=True)
    @classmethod
    def check_absence_of_empty_and_invalid_values(cls, values: Any) -> Any:
        """Check the model for empty and invalid values.

        Check that the `context` field contains `platform` and `revision` fields
        only if the `object.objectType` property is equal to `Activity`.
        """
        for field, value in list(values.items()):
            if value in [None, "", {}]:
                raise ValueError(f"{field}: invalid empty value")
            if isinstance(value, dict) and field != "extensions":
                cls.check_absence_of_empty_and_invalid_values(value)

        context = dict(values.get("context", {}))
        if context:
            platform = context.get("platform", {})
            revision = context.get("revision", {})
            object_type = dict(values["object"]).get("objectType", "Activity")
            if (platform or revision) and object_type != "Activity":
                raise ValueError(
                    "revision and platform properties can only be used if the "
                    "Statement's Object is an Activity"
                )
        return values



"""Swift data backend for Ralph."""

import logging
from functools import cached_property
from io import IOBase
from typing import Any, Iterable, Iterator, Optional, Tuple, Union
from uuid import uuid4

from pydantic import PositiveInt
from swiftclient.service import ClientException, Connection

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseOperationType,
    DataBackendStatus,
    Listable,
    Writable,
)
from ralph.backends.data.mixins import HistoryMixin
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now, parse_iterable_to_dict

logger = logging.getLogger(__name__)


class SwiftDataBackendSettings(BaseDataBackendSettings):
    """Swift data backend default configuration.

    Attributes:
        AUTH_URL (str): The authentication URL.
        USERNAME (str): The name of the openstack swift user.
        PASSWORD (str): The password of the openstack swift user.
        IDENTITY_API_VERSION (str): The keystone API version to authenticate to.
        TENANT_ID (str): The identifier of the tenant of the container.
        TENANT_NAME (str): The name of the tenant of the container.
        PROJECT_DOMAIN_NAME (str): The project domain name.
        REGION_NAME (str): The region where the container is.
        OBJECT_STORAGE_URL (str): The default storage URL.
        USER_DOMAIN_NAME (str): The user domain name.
        DEFAULT_CONTAINER (str): The default target container.
        LOCALE_ENCODING (str): The encoding used for reading/writing documents.
        READ_CHUNK_SIZE (str): The default chunk size for reading objects.
        WRITE_CHUNK_SIZE (str): The default chunk size for writing objects.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__SWIFT__"

    AUTH_URL: str = "https://auth.cloud.ovh.net/"
    DEFAULT_CONTAINER: Optional[str] = None
    IDENTITY_API_VERSION: str = "3"
    OBJECT_STORAGE_URL: Optional[str] = None
    PASSWORD: Optional[str] = None
    PROJECT_DOMAIN_NAME: str = "Default"
    READ_CHUNK_SIZE: int = 4096
    REGION_NAME: Optional[str] = None
    TENANT_ID: Optional[str] = None
    TENANT_NAME: Optional[str] = None
    USERNAME: Optional[str] = None
    USER_DOMAIN_NAME: str = "Default"
    WRITE_CHUNK_SIZE: int = 4096


class SwiftDataBackend(
    BaseDataBackend[SwiftDataBackendSettings, str],
    HistoryMixin,
    Writable,
    Listable,
):
    """SWIFT data backend."""

    name = "swift"
    default_operation_type = BaseOperationType.CREATE
    unsupported_operation_types = {
        BaseOperationType.APPEND,
        BaseOperationType.DELETE,
        BaseOperationType.UPDATE,
    }

    def __init__(self, settings: Optional[SwiftDataBackendSettings] = None):
        """Prepares the options for the SwiftService."""
        super().__init__(settings)
        self.default_container = self.settings.DEFAULT_CONTAINER
        self.locale_encoding = self.settings.LOCALE_ENCODING
        self._connection = None

    @cached_property
    def options(self) -> dict:
        """Return the required options for the Swift Connection."""
        return {
            "tenant_id": self.settings.TENANT_ID,
            "tenant_name": self.settings.TENANT_NAME,
            "project_domain_name": self.settings.PROJECT_DOMAIN_NAME,
            "region_name": self.settings.REGION_NAME,
            "object_storage_url": self.settings.OBJECT_STORAGE_URL,
            "user_domain_name": self.settings.USER_DOMAIN_NAME,
        }

    @property
    def connection(self) -> Connection:
        """Create a Swift Connection if it doesn't exist."""
        if not self._connection:
            self._connection = Connection(
                authurl=self.settings.AUTH_URL,
                user=self.settings.USERNAME,
                key=self.settings.PASSWORD,
                os_options=self.options,
                auth_version=self.settings.IDENTITY_API_VERSION,
            )
        return self._connection

    def status(self) -> DataBackendStatus:
        """Implement data backend checks (e.g. connection, cluster status).

        Return:
            DataBackendStatus: The status of the data backend.
        """
        try:
            self.connection.head_account()
        except ClientException as error:
            msg = "Unable to connect to the Swift account: %s"
            logger.error(msg, error.msg)
            return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List files for the target container.

        Args:
            target (str or None): The target container to list from.
                If `target` is `None`, the `default_container` will be used.
            details (bool): Get detailed object information instead of just names.
            new (bool): Given the history, list only not already read objects.

        Yield:
            str: The next object path. (If details is False)
            dict: The next object details. (If `details` is True.)

        Raise:
            BackendException: If a failure occurs.
        """
        if target is None:
            target = self.default_container

        archives_to_skip = set()
        if new:
            archives_to_skip = set(self.get_command_history(self.name, "read"))

        try:
            _, objects = self.connection.get_container(
                container=target, full_listing=True
            )
        except ClientException as error:
            msg = "Failed to list container %s: %s"
            logger.error(msg, target, error.msg)
            raise BackendException(msg % (target, error.msg)) from error

        for obj in objects:
            if new and obj in archives_to_skip:
                continue
            yield self._details(target, obj) if details else obj

    def read(  # noqa: PLR0913
        self,
        query: Optional[str] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read objects matching the `query` in the `target` container and yield them.

        Args:
            query (str): The query to select objects to read.
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            chunk_size (int or None): The number of records or bytes to read in one
                batch, depending on whether the records are dictionaries or bytes.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
                If the objects are dictionaries and `raw_output` is set to `True`, they
                are encoded as JSON.
                If the objects are bytes and `raw_output` is set to `False`, they are
                decoded as JSON by line.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            dict: If `raw_output` is False.
            bytes: If `raw_output` is True.

        Raise:
            BackendException: If a failure during the read operation occurs or
                during encoding records and `ignore_errors` is set to `False`.
            BackendParameterException: If a backend argument value is not valid.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_bytes(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> Iterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        target = target if target else self.default_container
        logger.info(
            "Getting object from container: %s (query: %s)",
            target,
            query,
        )
        resp_headers, content = self._get_object(target, query, chunk_size)
        while chunk := content.read(chunk_size):
            yield chunk
        # Archive read, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "action": "read",
                "id": f"{target}/{query}",
                "size": resp_headers["content-length"],
                "timestamp": now(),
            }
        )

    def _read_dicts(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        target = target if target else self.default_container
        logger.info("Getting object: %s from target container: %s", query, target)
        resp_headers, content = self._get_object(target, query, chunk_size)
        yield from parse_iterable_to_dict(content, ignore_errors)
        # Archive read, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "action": "read",
                "id": f"{target}/{query}",
                "size": resp_headers["content-length"],
                "timestamp": now(),
            }
        )

    def _get_object(
        self, container: Optional[str], obj: Optional[str], chunk_size: int
    ) -> Tuple[dict, Any]:
        """Validate container and obj and return Swift object wrapping the exception."""
        if not container:
            msg = "The target container is not set"
            logger.error(msg)
            raise BackendParameterException(msg)

        if not obj:
            msg = "The object query is not set"
            logger.error(msg)
            raise BackendParameterException(msg)

        try:
            return self.connection.get_object(
                container=container,
                obj=obj,
                resp_chunk_size=chunk_size,
            )
        except ClientException as error:
            msg = "Failed to read %s: %s"
            logger.error(msg, obj, error.msg)
            raise BackendException(msg % (obj, error.msg)) from error

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write `data` records to the `target` container and returns their count.

        Args:
            data (Iterable or IOBase): The data to write.
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            chunk_size (int or None): The chunk size when writing objects by batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, errors during decoding and encoding of
                records are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.

        Return:
            int: The number of written records.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If a backend argument value is not valid.
        """
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        return super()._write_dicts(
            data, target, chunk_size, ignore_errors, operation_type
        )

    def _write_bytes(  # noqa: PLR0913
        self,
        data: Iterable[bytes],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing bytes. See `self.write`."""
        counter = {"count": 0}
        data = self._count(data, counter)
        if not target:
            target = f"{self.default_container}/{now()}-{uuid4()}"
            msg = (
                "Target not specified; "
                "using default container with random object name: %s"
            )
            logger.info(msg, target)
        elif "/" not in target:
            target = f"{self.default_container}/{target}"
            msg = "Container not specified; using default container: %s"
            logger.info(msg, self.default_container)

        target_container, target_object = target.split("/", 1)
        if target_object in list(self.list(target=target_container)):
            msg = "%s already exists and overwrite is not allowed for operation %s"
            logger.error(msg, target_object, operation_type)
            raise BackendException(msg % (target_object, operation_type))

        try:
            self.connection.put_object(
                container=target_container,
                obj=target_object,
                contents=data,
                chunk_size=chunk_size,
            )
            resp = self.connection.head_object(
                container=target_container, obj=target_object
            )
        except ClientException as error:
            msg = "Failed to write to object %s: %s"
            logger.error(msg, target_object, error.msg)
            raise BackendException(msg % (target_object, error.msg)) from error

        count = counter["count"]
        logger.info("Successfully written %s statements to %s", count, target)

        # Archive written, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "action": "write",
                "operation_type": operation_type.value,
                "id": target,
                "size": resp["content-length"],
                "timestamp": now(),
            }
        )
        return count

    def close(self) -> None:
        """Close the Swift backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._connection:
            logger.warning("No backend client to close.")
            return

        try:
            self.connection.close()
        except ClientException as error:
            msg = "Failed to close Swift backend client: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    def _details(self, container: str, name: str):
        """Return `name` object details from `container`."""
        try:
            resp = self.connection.head_object(container=container, obj=name)
        except ClientException as error:
            msg = "Unable to retrieve details for object %s: %s"
            logger.error(msg, name, error.msg)
            raise BackendException(msg % (name, error.msg)) from error

        return {
            "name": name,
            "lastModified": resp["last-modified"],
            "size": resp["content-length"],
        }

    @staticmethod
    def _count(statements: Iterable, counter: dict) -> Iterator:
        """Count the elements in the `statements` Iterable and yield element."""
        for statement in statements:
            yield statement
            counter["count"] += 1



"""FileSystem LRS backend for Ralph."""

import logging
from datetime import datetime
from io import IOBase
from pathlib import Path
from typing import Iterable, List, Literal, Optional, Union
from uuid import UUID

from ralph.backends.data.base import BaseOperationType
from ralph.backends.data.fs import FSDataBackend, FSDataBackendSettings
from ralph.backends.lrs.base import (
    AgentParameters,
    BaseLRSBackend,
    BaseLRSBackendSettings,
    RalphStatementsQuery,
    StatementQueryResult,
)
from ralph.conf import BaseSettingsConfig

logger = logging.getLogger(__name__)


class FSLRSBackendSettings(BaseLRSBackendSettings, FSDataBackendSettings):
    """FileSystem LRS backend default configuration.

    Attributes:
        DEFAULT_LRS_FILE (str): The default LRS filename to store statements.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__LRS__FS__"

    DEFAULT_LRS_FILE: str = "fs_lrs.jsonl"


class FSLRSBackend(BaseLRSBackend[FSLRSBackendSettings], FSDataBackend):
    """FileSystem LRS Backend."""

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write data records to the target file and return their count.

        See `FSDataBackend.write`.
        """
        if target:
            target = str(Path(target) / Path(self.settings.DEFAULT_LRS_FILE))
        else:
            target = self.settings.DEFAULT_LRS_FILE
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""
        filters = []
        self._add_filter_by_id(filters, params.statement_id)
        self._add_filter_by_agent(filters, params.agent, params.related_agents)
        self._add_filter_by_authority(filters, params.authority)
        self._add_filter_by_verb(filters, params.verb)
        self._add_filter_by_activity(
            filters, params.activity, params.related_activities
        )
        self._add_filter_by_registration(filters, params.registration)
        self._add_filter_by_timestamp_since(filters, params.since)
        self._add_filter_by_timestamp_until(filters, params.until)
        self._add_filter_by_search_after(filters, params.search_after)

        limit = params.limit
        statements_count = 0
        search_after = None
        statements = []
        for statement in self.read(query=self.settings.DEFAULT_LRS_FILE, target=target):
            for query_filter in filters:
                if not query_filter(statement):
                    break
            else:
                statements.append(statement)
                statements_count += 1
                if limit and statements_count == limit:
                    search_after = statements[-1].get("id")
                    break

        if params.ascending:
            statements.reverse()
        return StatementQueryResult(
            statements=statements,
            pit_id=None,
            search_after=search_after,
        )

    def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> List:
        """Return the list of matching statement IDs from the database."""
        statement_ids = set(ids)
        statements = []
        for statement in self.read(query=self.settings.DEFAULT_LRS_FILE, target=target):
            if statement.get("id") in statement_ids:
                statements.append(statement)

        return statements

    @staticmethod
    def _add_filter_by_agent(
        filters: list, agent: Optional[AgentParameters], related: Optional[bool]
    ) -> None:
        """Add agent filters to `filters` if `agent` is set."""
        if not agent:
            return

        if not isinstance(agent, dict):
            agent = agent.dict()
        FSLRSBackend._add_filter_by_mbox(filters, agent.get("mbox", None), related)
        FSLRSBackend._add_filter_by_sha1sum(
            filters, agent.get("mbox_sha1sum", None), related
        )
        FSLRSBackend._add_filter_by_openid(filters, agent.get("openid", None), related)
        FSLRSBackend._add_filter_by_account(
            filters,
            agent.get("account__name", None),
            agent.get("account__home_page", None),
            related,
        )

    @staticmethod
    def _add_filter_by_authority(
        filters: list,
        authority: Optional[AgentParameters],
    ) -> None:
        """Add authority filters to `filters` if `authority` is set."""
        if not authority:
            return

        if not isinstance(authority, dict):
            authority = authority.dict()
        FSLRSBackend._add_filter_by_mbox(
            filters, authority.get("mbox", None), field="authority"
        )
        FSLRSBackend._add_filter_by_sha1sum(
            filters, authority.get("mbox_sha1sum", None), field="authority"
        )
        FSLRSBackend._add_filter_by_openid(
            filters, authority.get("openid", None), field="authority"
        )
        FSLRSBackend._add_filter_by_account(
            filters,
            authority.get("account__name", None),
            authority.get("account__home_page", None),
            field="authority",
        )

    @staticmethod
    def _add_filter_by_id(filters: list, statement_id: Optional[str]) -> None:
        """Add the `match_statement_id` filter if `statement_id` is set."""

        def match_statement_id(statement: dict) -> bool:
            """Return `True` if the statement has the given `statement_id`."""
            return statement.get("id") == statement_id

        if statement_id:
            filters.append(match_statement_id)

    @staticmethod
    def _get_related_agents(statement: dict) -> Iterable[dict]:
        yield statement.get("actor", {})
        yield statement.get("object", {})
        yield statement.get("authority", {})
        context = statement.get("context", {})
        yield context.get("instructor", {})
        yield context.get("team", {})

    @staticmethod
    def _add_filter_by_mbox(
        filters: list,
        mbox: Optional[str],
        related: Optional[bool] = False,
        field: Literal["actor", "authority"] = "actor",
    ) -> None:
        """Add the `match_mbox` filter if `mbox` is set."""

        def match_mbox(statement: dict) -> bool:
            """Return `True` if the statement has the given `actor.mbox`."""
            return statement.get(field, {}).get("mbox") == mbox

        def match_related_mbox(statement: dict) -> bool:
            """Return `True` if the statement has any agent matching `mbox`."""
            for agent in FSLRSBackend._get_related_agents(statement):
                if agent.get("mbox") == mbox:
                    return True

            statement_object = statement.get("object", {})
            if statement_object.get("objectType") == "SubStatement":
                return match_related_mbox(statement_object)
            return False

        if mbox:
            filters.append(match_related_mbox if related else match_mbox)

    @staticmethod
    def _add_filter_by_sha1sum(
        filters: list,
        sha1sum: Optional[str],
        related: Optional[bool] = False,
        field: Literal["actor", "authority"] = "actor",
    ) -> None:
        """Add the `match_sha1sum` filter if `sha1sum` is set."""

        def match_sha1sum(statement: dict) -> bool:
            """Return `True` if the statement has the given `actor.sha1sum`."""
            return statement.get(field, {}).get("mbox_sha1sum") == sha1sum

        def match_related_sha1sum(statement: dict) -> bool:
            """Return `True` if the statement has any agent matching `sha1sum`."""
            for agent in FSLRSBackend._get_related_agents(statement):
                if agent.get("mbox_sha1sum") == sha1sum:
                    return True

            statement_object = statement.get("object", {})
            if statement_object.get("objectType") == "SubStatement":
                return match_related_sha1sum(statement_object)
            return False

        if sha1sum:
            filters.append(match_related_sha1sum if related else match_sha1sum)

    @staticmethod
    def _add_filter_by_openid(
        filters: list,
        openid: Optional[str],
        related: Optional[bool] = False,
        field: Literal["actor", "authority"] = "actor",
    ) -> None:
        """Add the `match_openid` filter if `openid` is set."""

        def match_openid(statement: dict) -> bool:
            """Return `True` if the statement has the given `actor.openid`."""
            return statement.get(field, {}).get("openid") == openid

        def match_related_openid(statement: dict) -> bool:
            """Return `True` if the statement has any agent matching `openid`."""
            for agent in FSLRSBackend._get_related_agents(statement):
                if agent.get("openid") == openid:
                    return True

            statement_object = statement.get("object", {})
            if statement_object.get("objectType") == "SubStatement":
                return match_related_openid(statement_object)
            return False

        if openid:
            filters.append(match_related_openid if related else match_openid)

    @staticmethod
    def _add_filter_by_account(
        filters: list,
        name: Optional[str],
        home_page: Optional[str],
        related: Optional[bool] = False,
        field: Literal["actor", "authority"] = "actor",
    ) -> None:
        """Add the `match_account` filter if `name` or `home_page` is set."""

        def match_account(statement: dict) -> bool:
            """Return `True` if the statement has the given `actor.account`."""
            account = statement.get(field, {}).get("account", {})
            return account.get("name") == name and account.get("homePage") == home_page

        def match_related_account(statement: dict) -> bool:
            """Return `True` if the statement has any agent matching the account."""
            for agent in FSLRSBackend._get_related_agents(statement):
                account = agent.get("account", {})
                if account.get("name") == name and account.get("homePage") == home_page:
                    return True

            statement_object = statement.get("object", {})
            if statement_object.get("objectType") == "SubStatement":
                return match_related_account(statement_object)
            return False

        if name and home_page:
            filters.append(match_related_account if related else match_account)

    @staticmethod
    def _add_filter_by_verb(filters: list, verb_id: Optional[str]) -> None:
        """Add the `match_verb_id` filter if `verb_id` is set."""

        def match_verb_id(statement: dict) -> bool:
            """Return `True` if the statement has the given `verb.id`."""
            return statement.get("verb", {}).get("id") == verb_id

        if verb_id:
            filters.append(match_verb_id)

    @staticmethod
    def _add_filter_by_activity(
        filters: list, object_id: Optional[str], related: Optional[bool]
    ) -> None:
        """Add the `match_object_id` filter if `object_id` is set."""

        def match_object_id(statement: dict) -> bool:
            """Return `True` if the statement has the given `object.id`."""
            return statement.get("object", {}).get("id") == object_id

        def match_related_object_id(statement: dict) -> bool:
            """Return `True` if the statement has any object.id matching `object_id`."""
            statement_object = statement.get("object", {})
            if statement_object.get("id") == object_id:
                return True
            activities = statement.get("context", {}).get("contextActivities", {})
            for activity in activities.values():
                if isinstance(activity, dict):
                    if activity.get("id") == object_id:
                        return True
                else:
                    for sub_activity in activity:
                        if sub_activity.get("id") == object_id:
                            return True
            if statement_object.get("objectType") == "SubStatement":
                return match_related_object_id(statement_object)

            return False

        if object_id:
            filters.append(match_related_object_id if related else match_object_id)

    def _add_filter_by_timestamp_since(
        self, filters: list, timestamp: Optional[datetime]
    ) -> None:
        """Add the `match_since` filter if `timestamp` is set."""
        if isinstance(timestamp, str):
            timestamp = datetime.fromisoformat(timestamp)

        def match_since(statement: dict) -> bool:
            """Return `True` if the statement was created after `timestamp`."""
            try:
                statement_timestamp = datetime.fromisoformat(statement.get("timestamp"))
            except (TypeError, ValueError) as error:
                msg = "Statement with id=%s contains unparsable timestamp=%s"
                logger.debug(msg, statement.get("id"), error)
                return False
            return statement_timestamp > timestamp

        if timestamp:
            filters.append(match_since)

    def _add_filter_by_timestamp_until(
        self, filters: list, timestamp: Optional[datetime]
    ) -> None:
        """Add the `match_until` function if `timestamp` is set."""
        if isinstance(timestamp, str):
            timestamp = datetime.fromisoformat(timestamp)

        def match_until(statement: dict) -> bool:
            """Return `True` if the statement was created before `timestamp`."""
            try:
                statement_timestamp = datetime.fromisoformat(statement.get("timestamp"))
            except (TypeError, ValueError) as error:
                msg = "Statement with id=%s contains unparsable timestamp=%s"
                logger.debug(msg, statement.get("id"), error)
                return False
            return statement_timestamp <= timestamp

        if timestamp:
            filters.append(match_until)

    @staticmethod
    def _add_filter_by_search_after(filters: list, search_after: Optional[str]) -> None:
        """Add the `match_search_after` filter if `search_after` is set."""
        search_after_state = {"state": False}

        def match_search_after(statement: dict) -> bool:
            """Return `True` if the statement was created after `search_after`."""
            if search_after_state["state"]:
                return True
            if statement.get("id") == search_after:
                search_after_state["state"] = True
            return False

        if search_after:
            filters.append(match_search_after)

    @staticmethod
    def _add_filter_by_registration(
        filters: list, registration: Optional[UUID]
    ) -> None:
        """Add the `match_registration` filter if `registration` is set."""
        registration_str = str(registration)

        def match_registration(statement: dict) -> bool:
            """Return `True` if the statement has the given `context.registration`."""
            return statement.get("context", {}).get("registration") == registration_str

        if registration:
            filters.append(match_registration)



"""Video xAPI events result fields definitions."""

import sys
from datetime import timedelta
from typing import Optional

from pydantic import Field, NonNegativeFloat

from ..base.results import BaseXapiResult
from ..concepts.constants.video import (
    CONTEXT_EXTENSION_CC_ENABLED,
    CONTEXT_EXTENSION_PLAYED_SEGMENTS,
    RESULT_EXTENSION_PROGRESS,
    RESULT_EXTENSION_TIME,
    RESULT_EXTENSION_TIME_FROM,
    RESULT_EXTENSION_TIME_TO,
)
from ..config import BaseExtensionModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class VideoResultExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for video `result`.`extensions` property.

    Attributes:
        playedSegments (str): Consists of parts of the video the actor watched during
            current registration in chronological order (for example,
            "0[.]5[,]12[.]22[,]15[.]55[,]55[.]99.33[,]99.33").
        time (float): Consists of the video time code when the event was emitted.
    """

    time: NonNegativeFloat = Field(alias=RESULT_EXTENSION_TIME)
    playedSegments: Optional[str] = Field(alias=CONTEXT_EXTENSION_PLAYED_SEGMENTS)


class VideoPausedResultExtensions(VideoResultExtensions):
    """Pydantic model for video paused `result`.`extensions` property.

    Attributes:
        progress (float): Consists of the ratio of media consumed by the actor.
    """

    progress: Optional[NonNegativeFloat] = Field(alias=RESULT_EXTENSION_PROGRESS)


class VideoSeekedResultExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for video seeked `result`.`extensions` property.

    Attributes:
        timeFrom (float): Consists of the point in time the actor changed from in a
            media object during a seek operation.
        timeTo (float): Consists of the point in time the actor changed to in a media
            object during a seek operation.
    """

    timeFrom: NonNegativeFloat = Field(alias=RESULT_EXTENSION_TIME_FROM)
    timeTo: NonNegativeFloat = Field(alias=RESULT_EXTENSION_TIME_TO)


class VideoCompletedResultExtensions(VideoResultExtensions):
    """Pydantic model for video completed `result`.`extensions` property.

    Attributes:
        progress (float): Consists of the percentage of media consumed by the actor.
    """

    progress: NonNegativeFloat = Field(alias=RESULT_EXTENSION_PROGRESS)


class VideoTerminatedResultExtensions(VideoResultExtensions):
    """Pydantic model for video terminated `result`.`extensions` property.

    Attributes:
        progress (float): Consists of the percentage of media consumed by the actor.
    """

    progress: NonNegativeFloat = Field(alias=RESULT_EXTENSION_PROGRESS)


class VideoEnableClosedCaptioningResultExtensions(VideoResultExtensions):
    """Pydantic model for video enable closed captioning `result`.`extensions` property.

    Attributes:
        ccEnabled (bool): Indicates whether subtitles are enabled.
    """

    ccEnabled: bool = Field(alias=CONTEXT_EXTENSION_CC_ENABLED)


class VideoPlayedResult(BaseXapiResult):
    """Pydantic model for video played `result` property.

    Attributes:
        extensions (dict): See VideoResultExtensions.
    """

    extensions: VideoResultExtensions


class VideoPausedResult(BaseXapiResult):
    """Pydantic model for video paused `result` property.

    Attributes:
        extensions (dict): See VideoPausedResultExtensions.
    """

    extensions: VideoPausedResultExtensions


class VideoSeekedResult(BaseXapiResult):
    """Pydantic model for video seeked `result` property.

    Attributes:
        extensions (dict): See VideoSeekedResultExtensions.
    """

    extensions: VideoSeekedResultExtensions


class VideoCompletedResult(BaseXapiResult):
    """Pydantic model for video completed `result` property.

    Attributes:
        extensions (dict): See VideoCompletedResultExtensions.
        completion (bool): Consists of the value `True`.
        duration (str): Consists of the total time spent consuming the video under
            current registration.
    """

    extensions: VideoCompletedResultExtensions
    completion: Optional[Literal[True]]
    duration: Optional[timedelta]


class VideoTerminatedResult(BaseXapiResult):
    """Pydantic model for video terminated `result` property.

    Attributes:
        extensions (dict): See VideoTerminatedResultExtensions.
    """

    extensions: VideoTerminatedResultExtensions


class VideoEnableClosedCaptioningResult(BaseXapiResult):
    """Pydantic model for video enable closed captioning `result` property.

    Attributes:
        extensions (dict): See VideoEnableClosedCaptioningResultExtensions.
    """

    extensions: VideoEnableClosedCaptioningResultExtensions


class VideoVolumeChangeInteractionResult(BaseXapiResult):
    """Pydantic model for video volume change interaction `result` property.

    Attributes:
        extensions (dict): See VideoResultExtensions.
    """

    extensions: VideoResultExtensions


class VideoScreenChangeInteractionResult(BaseXapiResult):
    """Pydantic model for video screen change interaction `result` property.

    Attributes:
        extensions (dict): See VideoResultExtensions.
    """

    extensions: VideoResultExtensions



"""`Virtual classroom` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Virtual classroom


class VirtualClassroomActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for virtual classroom `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom`.
    """

    type: Literal[
        "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"
    ] = "https://w3id.org/xapi/virtual-classroom/activity-types/virtual-classroom"


class VirtualClassroomActivity(BaseXapiActivity):
    """Pydantic model for virtual classroom `Activity` type.

    Attributes:
        definition (dict): See VirtualClassroomActivityDefinition.
    """

    definition: VirtualClassroomActivityDefinition = (
        VirtualClassroomActivityDefinition()
    )
